from __future__ import annotations

import pytest

from exo import proc
from exo.platforms.x86 import *

from exo.stdlib.scheduling import *
from exo.stdlib.halide_scheduling_ops import *


@proc
def blur1d_compute_root(n: size, consumer: i8[n], inp: i8[n + 6]):
    producer: i8[n + 1]
    for i in seq(0, n + 1):
        producer[i] = (
            inp[i] + inp[i + 1] + inp[i + 2] + inp[i + 3] + inp[i + 4] + inp[i + 5]
        ) / 6.0

    for i in seq(0, n):
        consumer[i] = (producer[i] + producer[i + 1]) / 2.0


def test_schedule_blur1d(golden):
    p = blur1d_compute_root
    procs = []

    loop = p.find_loop("i #1")
    producer_assign = p.find("producer = _")
    p = compute_at(p, producer_assign, loop)
    p = rename(p, "blur1d_compute_at_store_root")
    procs.append(p)

    loop = p.find_loop("i")
    producer_alloc = p.find("producer : _")
    p = store_at(p, producer_alloc, loop)
    p = rename(p, "blur1d_compute_at")
    procs.append(p)

    p = unroll_loop(p, "ii")
    for i in range(2):
        p = inline_assign(p, p.find("consumer[_] = _").prev())
    p = delete_buffer(p, "producer: _")
    p = simplify(p)
    p = rename(p, "blur1d_inline")
    procs.append(p)

    print("\n\n".join([str(p) for p in procs]))
    assert "\n\n".join([str(p) for p in procs]) == golden


@proc
def blur2d_compute_root(n: size, consumer: i8[n, n], sin: i8[n + 1, n + 1]):
    assert n % 4 == 0
    producer: i8[n + 1, n + 1]
    for i in seq(0, n + 1):
        for j in seq(0, n + 1):
            producer[i, j] = sin[
                i, j
            ]  # just a placeholder since sine can't evalute on index exprs

    for i in seq(0, n):
        for j in seq(0, n):
            consumer[i, j] = (
                producer[i, j]
                + producer[i, j + 1]
                + producer[i + 1, j]
                + producer[i + 1, j + 1]
            ) / 4.0


def test_schedule_blur2d(golden):
    p = blur2d_compute_root
    procs = []

    producer_assign = p.find("producer = _")
    producer_alloc = p.find("producer : _")

    p = compute_at(p, producer_assign, p.find_loop("i #1"))
    p = rename(p, "blur2d_compute_at_i_store_root")
    procs.append(p)

    p = blur2d_compute_root
    p = compute_at(p, producer_assign, p.find_loop("j #1"))
    p = rename(p, "blur2d_compute_at_j_store_root")
    procs.append(p)

    p = blur2d_compute_root
    p = compute_at(p, producer_assign, p.find_loop("i #1"))
    p = store_at(p, producer_alloc, p.find_loop("i"))
    p = rename(p, "blur2d_compute_at_i")
    procs.append(p)

    p = blur2d_compute_root
    p = compute_at(p, producer_assign, p.find_loop("j #1"))
    p = store_at(p, producer_alloc, p.find_loop("i"))
    p = simplify(p)
    p = rename(p, "blur2d_compute_at_j_store_at_i")
    procs.append(p)

    p = blur2d_compute_root
    p = compute_at(p, producer_assign, p.find_loop("j #1"))
    p = store_at(p, producer_alloc, p.find_loop("j"))
    p = unroll_loop(p, "ii")
    p = unroll_loop(p, "ji")
    for i in range(4):
        p = inline_assign(p, p.find("consumer[_] = _").prev())
    p = delete_buffer(p, "producer: _")
    p = rename(p, "blur2d_inline")
    procs.append(p)

    assert "\n\n".join([str(p) for p in procs]) == golden


@pytest.mark.slow
def test_schedule_tiled_blur2d(golden):
    compute_root = blur2d_compute_root
    procs = []

    p_tiled = tile(
        compute_root,
        compute_root.find_loop("i #1"),
        compute_root.find_loop("j #1"),
        ["i", "ii"],
        ["j", "ji"],
        4,
        4,
        perfect=True,
    )

    producer_assign = p_tiled.find("producer = _")
    producer_alloc = p_tiled.find("producer : _")

    p = p_tiled
    p = rename(p, "blur2d_tiled")
    procs.append(p)

    p = p_tiled
    p = compute_at(p, producer_assign, p.find_loop("i #1"))
    p = rename(p, "blur2d_tiled_compute_at_i")
    procs.append(p)

    p = p_tiled
    p = compute_at(p, producer_assign, p.find_loop("j #1"))
    p = rename(p, "blur2d_tiled_compute_at_j")
    procs.append(p)

    p = p_tiled
    p = compute_at(p, producer_assign, p.find_loop("ii"))
    p = rename(p, "blur2d_tiled_compute_at_ii")
    procs.append(p)

    p = p_tiled
    p = compute_at(p, producer_assign, p.find_loop("ji"))
    p = rename(p, "blur2d_tiled_compute_at_ji")
    procs.append(p)

    p = p_tiled
    p = compute_at(p, producer_assign, p.find_loop("ji"))
    p = store_at(p, producer_alloc, p.find_loop("ji"))
    p = rename(p, "blur2d_tiled_compute_at_and_store_at_ji")
    procs.append(p)

    assert "\n\n".join([str(p) for p in procs]) == golden



from __future__ import annotations

import inspect
import re
from typing import Optional, Iterable

import exo.pyparser as pyparser
from exo.LoopIR import LoopIR, PAST

# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Pattern Matching Errors
from exo.internal_cursors import Cursor, Node, Block


class PatternMatchError(Exception):
    pass


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# General Pattern-Matching / Pointing Mechanism

"""
We will use <pattern-string>s as a way to point at AST nodes.

A <pattern-string> has the following form:

<pattern-string> ::= <pattern> #<num>
                   | <pattern>
<pattern>        ::= ... -- a UAST statement or expression
                         -- potentially involving one or more holes
                         -- where a hole is written `_`
                         -- specified by LoopIR.PAST
"""


def get_match_no(pattern_str: str) -> Optional[int]:
    """
    Search for a trailing # sign in a pattern string and return the following
    number, or None if it does no # sign exists. Uses `int` to parse the number,
    so is not sensitive to spaces
    >>> get_match_no('foo #34')
    34
    >>> get_match_no('baz # 42 ')
    42
    >>> get_match_no('foo') is None
    True
    >>> get_match_no('foo #bar')
    Traceback (most recent call last):
      ...
    ValueError: invalid literal for int() with base 10: 'bar'
    """
    if (pos := pattern_str.rfind("#")) == -1:
        return None
    return int(pattern_str[pos + 1 :])


def match_pattern(
    context: Cursor,
    pattern_str: str,
    call_depth=0,
    default_match_no=None,
    use_sym_id=False,
):
    """
    If [default_match_no] is None, then all matches are returned

    If [use_sym_id] is True, all symbols matchesrare additionally checked
    against their unique id, rather than just matching against the name.
    """
    assert isinstance(context, Cursor), f"Expected Cursor, got {type(context)}"

    # break-down pattern_str for possible #<num> post-fix
    if match := re.search(r"^([^#]+)#(\d+)\s*$", pattern_str):
        pattern_str = match[1]
        match_no = int(match[2]) if match[2] is not None else None
    else:
        match_no = default_match_no  # None means match-all

    # get source location where this is getting called from
    caller = inspect.getframeinfo(inspect.stack()[call_depth + 1][0])

    # parse the pattern we're going to use to match
    p_ast = pyparser.pattern(
        pattern_str, filename=caller.filename, lineno=caller.lineno
    )

    # do the pattern match, to find the nodes in ast
    return PatternMatch().find(context, p_ast, match_no=match_no, use_sym_id=use_sym_id)


_PAST_to_LoopIR = {
    # list of stmts
    list: list,
    #
    PAST.Assign: [LoopIR.Assign],
    PAST.Reduce: [LoopIR.Reduce],
    PAST.Pass: [LoopIR.Pass],
    PAST.If: [LoopIR.If],
    PAST.For: [LoopIR.For],
    PAST.Alloc: [LoopIR.Alloc],
    PAST.Call: [LoopIR.Call],
    PAST.WriteConfig: [LoopIR.WriteConfig],
    PAST.S_Hole: None,
    #
    PAST.Read: [LoopIR.Read],
    PAST.StrideExpr: [LoopIR.StrideExpr],
    PAST.Const: [LoopIR.Const],
    PAST.USub: [LoopIR.USub],
    PAST.BinOp: [LoopIR.BinOp],
    PAST.BuiltIn: [LoopIR.BuiltIn],
    PAST.ReadConfig: [LoopIR.ReadConfig],
    PAST.E_Hole: None,
}


class _MatchComplete(Exception):
    pass


class PatternMatch:
    def __init__(self):
        self._match_no = None
        self._results = []
        self._use_sym_id = False

    def find(self, cur, pat, match_no=None, use_sym_id=False):
        self._match_no = match_no
        self._results = []
        self._use_sym_id = use_sym_id

        # prevent the top level of a pattern being just a hole
        if isinstance(pat, PAST.E_Hole):
            raise PatternMatchError("pattern match on 'anything' unsupported")
        elif isinstance(pat, list) and all(isinstance(p, PAST.S_Hole) for p in pat):
            raise PatternMatchError("pattern match on 'anything' unsupported")

        try:
            if isinstance(pat, list):
                assert len(pat) > 0
                self.find_stmts(pat, cur)
            else:
                assert isinstance(pat, PAST.expr)
                self.find_expr(pat, cur)
        except _MatchComplete:
            pass

        return self._results

    def _add_result(self, result):
        assert isinstance(result, (Node, Block))

        if self._match_no is None:
            self._results.append(result)
            return

        i = self._match_no
        self._match_no -= 1
        if i == 0:
            self._results.append(result)
            raise _MatchComplete()

    ## -------------------
    ##  finding methods

    def find_expr(self, pat, cur):
        # try to match
        if self.match_e(pat, cur._node):
            self._add_result(cur)

        for child in _children(cur):
            self.find_expr(pat, child)

    def find_stmts(self, pats, cur: Node):
        if isinstance(cur._node, LoopIR.proc):
            return self.find_stmts_in_block(pats, cur.body())

        return self.find_stmts_in_block(pats, cur.as_block())

    def find_stmts_in_block(self, pats, curs: Block):
        # may encounter empty statement blocks, which we should ignore
        if len(curs) == 0:
            return

        # try to match a prefix of this sequence of statements
        if m := self.match_stmts(pats, curs):
            self._add_result(m)

        # if we need to look for more matches, recurse structurally ...

        # first, look for any subsequences of statements in the first
        # statement of the sequence `stmts`
        if isinstance(curs[0]._node, LoopIR.If):
            self.find_stmts_in_block(pats, curs[0].body())
            self.find_stmts_in_block(pats, curs[0].orelse())
        elif isinstance(curs[0]._node, LoopIR.For):
            self.find_stmts_in_block(pats, curs[0].body())
        else:
            pass  # other forms of statement do not contain stmt blocks

        # second, recurse on the tail of this sequence...
        self.find_stmts_in_block(pats, curs[1:])

    ## -------------------
    ##  matching methods

    def match_stmts(self, pats, cur):
        i, j = 0, 0
        while i < len(pats) and j < len(cur):
            if isinstance(pats[i], PAST.S_Hole):
                if i + 1 == len(pats):
                    return cur  # No lookahead, guaranteed match
                if self.match_stmt(pats[i + 1], cur[j]):
                    i += 2  # Lookahead matches, skip hole and lookahead
            elif self.match_stmt(pats[i], cur[j]):
                i += 1
            else:
                return None
            j += 1

        # Return the matched portion on success
        return cur[:j] if i == len(pats) else None

    def match_stmt(self, pat, cur):
        assert not isinstance(pat, PAST.S_Hole), "holes must be handled in match_stmts"

        # first ensure that the pattern and statement
        # are the same constructor

        stmt = cur._node

        if not isinstance(
            stmt, (LoopIR.WindowStmt,) + tuple(_PAST_to_LoopIR[type(pat)])
        ):
            return False

        # then handle each constructor as a structural case
        if isinstance(stmt, (LoopIR.Assign, LoopIR.Reduce)):
            return (
                self.match_name(pat.name, stmt.name)
                and all(self.match_e(pi, si) for pi, si in zip(pat.idx, stmt.idx))
                and self.match_e(pat.rhs, stmt.rhs)
            )
        elif isinstance(stmt, LoopIR.WindowStmt):
            if isinstance(pat, PAST.Assign):
                return (
                    self.match_name(pat.name, stmt.name)
                    and pat.idx == []
                    and self.match_e(pat.rhs, stmt.rhs)
                )
            else:
                return False
        elif isinstance(stmt, LoopIR.Pass):
            return True
        elif isinstance(stmt, LoopIR.If):
            return (
                self.match_e(pat.cond, stmt.cond)
                and self.match_stmts(pat.body, cur.body()) is not None
                and self.match_stmts(pat.orelse, cur.orelse()) is not None
            )
        elif isinstance(stmt, LoopIR.For):
            return (
                self.match_name(pat.iter, stmt.iter)
                and self.match_e(pat.lo, stmt.lo)
                and self.match_e(pat.hi, stmt.hi)
                and self.match_stmts(pat.body, cur.body()) is not None
            )
        elif isinstance(stmt, LoopIR.Alloc):
            if isinstance(stmt.type, LoopIR.Tensor):
                return all(
                    self.match_e(pi, si) for pi, si in zip(pat.sizes, stmt.type.hi)
                ) and self.match_name(pat.name, stmt.name)
            else:  # scalar
                return self.match_name(pat.name, stmt.name)
        elif isinstance(stmt, LoopIR.Call):
            return self.match_name(pat.f, stmt.f.name)
        elif isinstance(stmt, LoopIR.WriteConfig):
            return self.match_name(stmt.config.name(), pat.config) and self.match_name(
                stmt.field, pat.field
            )
        else:
            assert False, f"bad case: {type(stmt)}"

    def match_e(self, pat, e):
        # expression holes can match anything
        # and we don't have to worry about Kleene-Star behavior
        if isinstance(pat, PAST.E_Hole):
            return True

        # Special case: -3 can be parsed as USub(Const(3))... it should match Const(-3)
        if (
            isinstance(pat, PAST.USub)
            and isinstance(pat.arg, PAST.Const)
            and isinstance(e, LoopIR.Const)
        ):
            pat = pat.arg.update(val=-pat.arg.val)

        # first ensure that the pattern and statement
        # are the same constructor
        if not isinstance(e, (LoopIR.WindowExpr,) + tuple(_PAST_to_LoopIR[type(pat)])):
            return False

        if isinstance(e, LoopIR.Read):
            return self.match_name(pat.name, e.name) and all(
                self.match_e(pi, si) for pi, si in zip(pat.idx, e.idx)
            )
        elif isinstance(e, LoopIR.WindowExpr):
            if isinstance(pat, PAST.Read):
                # TODO: Should we be able to handle window slicing matching? Nah..
                if len(pat.idx) != 1 or not isinstance(pat.idx[0], PAST.E_Hole):
                    return False
                return self.match_name(pat.name, e.name)
            else:
                return False
        elif isinstance(e, LoopIR.Const):
            return pat.val == e.val
        elif isinstance(e, LoopIR.BinOp):
            # TODO: do we need to handle associativity? (a + b) + c vs a + (b + c)?
            return (
                pat.op == e.op
                and self.match_e(pat.lhs, e.lhs)
                and self.match_e(pat.rhs, e.rhs)
            )
        elif isinstance(e, LoopIR.USub):
            return self.match_e(pat.arg, e.arg)
        elif isinstance(e, LoopIR.BuiltIn):
            return pat.f is e.f and all(
                self.match_e(pa, sa) for pa, sa in zip(pat.args, e.args)
            )
        elif isinstance(e, LoopIR.ReadConfig):
            return pat.config == e.config.name() and pat.field == e.field
        elif isinstance(e, LoopIR.StrideExpr):
            return self.match_name(pat.name, e.name) and (
                pat.dim == e.dim or not bool(pat.dim)
            )
        else:
            assert False, "bad case"

    def match_name(self, pat_nm, ir_sym):
        # We use repr(sym) as a way of checking both the Sym name and id
        ir_sym = repr(ir_sym) if self._use_sym_id else str(ir_sym)
        return pat_nm == "_" or pat_nm == ir_sym


def _children(cur) -> Iterable[Node]:
    n = cur._node
    # Top-level proc
    if isinstance(n, LoopIR.proc):
        yield from _children_from_attrs(cur, n, "body")
    # Statements
    elif isinstance(n, (LoopIR.Assign, LoopIR.Reduce)):
        yield from _children_from_attrs(cur, n, "idx", "rhs")
    elif isinstance(n, (LoopIR.WriteConfig, LoopIR.WindowStmt)):
        yield from _children_from_attrs(cur, n, "rhs")
    elif isinstance(n, (LoopIR.Pass, LoopIR.Alloc, LoopIR.Free)):
        yield from []
    elif isinstance(n, LoopIR.If):
        yield from _children_from_attrs(cur, n, "cond", "body", "orelse")
    elif isinstance(n, LoopIR.For):
        yield from _children_from_attrs(cur, n, "lo", "hi", "body")
    elif isinstance(n, LoopIR.Call):
        yield from _children_from_attrs(cur, n, "args")
    # Expressions
    elif isinstance(n, LoopIR.Read):
        yield from _children_from_attrs(cur, n, "idx")
    elif isinstance(n, LoopIR.WindowExpr):
        yield from _children_from_attrs(cur, n, "idx")
    elif isinstance(n, LoopIR.Interval):
        yield from _children_from_attrs(cur, n, "lo", "hi")
    elif isinstance(n, LoopIR.Point):
        yield from _children_from_attrs(cur, n, "pt")
    elif isinstance(
        n,
        (
            LoopIR.Const,
            LoopIR.StrideExpr,
            LoopIR.ReadConfig,
        ),
    ):
        yield from []
    elif isinstance(n, LoopIR.USub):
        yield from _children_from_attrs(cur, n, "arg")
    elif isinstance(n, LoopIR.BinOp):
        yield from _children_from_attrs(cur, n, "lhs", "rhs")
    elif isinstance(n, LoopIR.BuiltIn):
        yield from _children_from_attrs(cur, n, "args")
    else:
        assert False, f"case {type(n)} unsupported"


def _children_from_attrs(cur, n, *args) -> Iterable[Node]:
    for attr in args:
        children = getattr(n, attr)
        if isinstance(children, list):
            for i in range(len(children)):
                yield cur._child_node(attr, i)
        else:
            yield cur._child_node(attr, None)



from __future__ import annotations

import pytest

from exo import proc, DRAM, config, instr
from exo.libs.memories import GEMM_SCRATCH
from exo.stdlib.scheduling import *


# ------- Configuration tests ---------


@pytest.mark.skip()
def test_config_size():
    def new_config_ld():
        @config
        class ConfigLoad:
            size: size

        return ConfigLoad

    ConfigLoad = new_config_ld()

    @proc
    def do_ld_i8(n: size):
        assert n == ConfigLoad.size
        pass

    # @proc
    # def config_ld_i8(
    #    config_n : size
    # ):
    #    ConfigLoad.size = config_n

    @proc
    def foo(N: size):
        # config_ld_i8(N)
        ConfigLoad.size = N
        do_ld_i8(N)

    print(foo)


@pytest.mark.skip()
def test_config_stride():
    def new_config_ld():
        @config
        class ConfigLoad:
            src_stride: stride

        return ConfigLoad

    ConfigLoad = new_config_ld()

    @proc
    def do_ld_i8(n: size, src: [i8][n] @ DRAM):
        assert stride(src, 0) == ConfigLoad.src_stride
        pass

    @proc
    def config_ld_i8(src_stride: stride):
        ConfigLoad.src_stride = src_stride

    @proc
    def foo(N: size, A: i8[N]):
        config_ld_i8(stride(A, 0))
        do_ld_i8()

    print(foo)


def new_config_f32():
    @config
    class ConfigAB:
        a: f32
        b: f32

    return ConfigAB


def new_control_config():
    @config
    class ConfigControl:
        i: index
        s: stride
        b: bool

    return ConfigControl


def test_basic_config(golden):
    ConfigAB = new_config_f32()

    @proc
    def foo(x: f32):
        ConfigAB.a = 32.0
        x = ConfigAB.a

    assert str(foo) == golden


def test_write_loop_const_number(golden):
    ConfigAB = new_config_f32()

    @proc
    def foo(n: size):
        for i in seq(0, n):
            ConfigAB.a = 0.0

    assert str(foo) == golden


def test_write_loop_builtin(golden):
    ConfigAB = new_config_f32()

    @proc
    def foo(n: size):
        for i in seq(0, n):
            ConfigAB.a = sin(1.0)

    assert str(foo) == golden


# Config loop dependency tests
@pytest.mark.skip()
def test_write_loop_varying():
    ConfigAB = new_config_f32()
    with pytest.raises(TypeError, match="The value written to config variable"):

        @proc
        def foo(n: size, A: f32[n]):
            for i in seq(0, n):
                ConfigAB.a = A[i]


@pytest.mark.skip()
def test_write_loop_varying_indirect():
    ConfigAB = new_config_f32()
    with pytest.raises(TypeError, match="The value written to config variable"):

        @proc
        def foo(n: size, A: f32[n]):
            for i in seq(0, n):
                a: f32
                a = A[i]
                ConfigAB.a = a


# NOTE: The following test documents current behavior
#       but it would be very reasonable to make this test
#       non-failing
# Fix is to improve the dependency analysis
@pytest.mark.skip()
def test_write_loop_syntax_check_fail():
    CTRL = new_control_config()

    with pytest.raises(TypeError, match="depends on the loop iteration variable"):

        @proc
        def foo(n: size):
            for i in seq(0, n):
                CTRL.i = i - i


def test_write_all_control(golden):
    CTRL = new_control_config()

    @proc
    def set_all(i: index, s: stride, b: bool):
        CTRL.i = i
        CTRL.s = s
        CTRL.b = b

    assert str(set_all) == golden


# Should the following succeed or fail?
# I think it probably should succeed
def test_loop_complex_guards(golden):
    CTRL = new_control_config()

    @proc
    def foo(n: size):
        for i in seq(0, n):
            if CTRL.i == 3:
                CTRL.i = 4
            if n == n - 1:
                CTRL.i = 3

    assert str(foo) == golden


@pytest.mark.skip()
def test_loop_circular_guards():
    CTRL = new_control_config()

    with pytest.raises(TypeError, match="TODO: Need to determine which error"):

        @proc
        def foo(n: size):
            for i in seq(0, n):
                if CTRL.i == 3:
                    CTRL.i = 4
                elif CTRL.i == 4:
                    CTRL.i = 3


# NOTE: I don't think this should work necessarily
@pytest.mark.skip()  # This should work
def test_config_write7():
    ConfigAB = new_config_f32()

    @proc
    def foo(n: size):
        a: f32
        for i in seq(0, n):
            ConfigAB.a = 3.0
            a = ConfigAB.a


def new_config_ld():
    @config
    class ConfigLoad:
        scale: f32
        src_stride: stride

    return ConfigLoad


def test_stride_with_config(golden):
    ConfigLoad = new_config_ld()

    @proc
    def bar(n: size, src: [i8][n]):
        assert stride(src, 0) == ConfigLoad.src_stride
        pass

    @proc
    def foo(n: size, src: [i8][n]):
        assert stride(src, 0) == ConfigLoad.src_stride
        bar(n, src)

    assert f"{bar}\n{foo}" == golden


def test_config_write(golden):
    @config
    class Config:
        tmp: f32

    @proc
    def foo():
        tmp: f32
        tmp = 0.0

    bar = write_config(foo, foo.find("tmp = _").after(), Config, "tmp", "tmp")
    assert str(bar) == golden


def test_config_bind(golden):
    ConfigLoad = new_config_ld()

    @proc
    def foo(scale: f32):
        for i in seq(0, 10):
            tmp: f32
            tmp = 0.0
            tmp = tmp * scale

    foo = bind_config(foo, "scale", ConfigLoad, "scale")

    assert str(foo) == golden


def test_config_fission(golden):
    ConfigLoad = new_config_ld()

    @proc
    def foo(scale: f32, n: size, m: size, A: f32[n, m]):
        for i in seq(0, n):
            for j in seq(0, m):
                ConfigLoad.scale = scale
                tmp: f32
                tmp = A[i, j]
                tmp = tmp * ConfigLoad.scale

    foo = autofission(foo, foo.find("ConfigLoad.scale = _").after(), n_lifts=2)

    assert str(foo) == golden


def test_ld(golden):
    ConfigLoad = new_config_ld()

    _gemm_config_ld_i8 = (
        "gemmini_extended3_config_ld({src_stride}, " + "{scale}[0], 0, 0);\n"
    )

    @instr(_gemm_config_ld_i8)
    def config_ld_i8(scale: f32, src_stride: stride):
        ConfigLoad.scale = scale
        ConfigLoad.src_stride = src_stride

    _gemm_do_ld_i8 = (
        "gemmini_extended_mvin( {src}.data, " + "((uint64_t) {dst}.data), {m}, {n} );"
    )

    @instr(_gemm_do_ld_i8)
    def do_ld_i8(
        n: size,
        m: size,
        src: i8[n, m] @ DRAM,
        dst: i8[n, 16] @ GEMM_SCRATCH,
    ):
        assert n <= 16
        assert m <= 16
        assert stride(src, 1) == 1
        assert stride(dst, 0) == 16
        assert stride(dst, 1) == 1
        assert stride(src, 0) == ConfigLoad.src_stride

        for i in seq(0, n):
            for j in seq(0, m):
                tmp: f32
                tmp = src[i, j]
                tmp = tmp * ConfigLoad.scale
                dst[i, j] = tmp

    _gemm_ld_i8 = (
        "gemmini_extended3_config_ld({stride(src, 0)}, "
        + "{scale}[0], 0, 0);\n"
        + "gemmini_extended_mvin( {src}.data, "
        + "((uint64_t) {dst}.data), {m}, {n} );"
    )

    @instr(_gemm_ld_i8)
    def ld_i8(
        n: size,
        m: size,
        scale: f32,
        src: i8[n, m] @ DRAM,
        dst: i8[n, 16] @ GEMM_SCRATCH,
    ):
        assert n <= 16
        assert m <= 16
        assert stride(src, 1) == 1
        assert stride(dst, 0) == 16
        assert stride(dst, 1) == 1

        for i in seq(0, n):
            for j in seq(0, m):
                tmp: f32
                tmp = src[i, j]
                tmp = tmp * scale
                dst[i, j] = tmp

    ld_i8 = bind_config(ld_i8, "scale", ConfigLoad, "scale")
    ld_i8 = reorder_stmts(ld_i8, "tmp = src[_] ; ConfigLoad.scale = _")
    ld_i8 = reorder_stmts(ld_i8, "tmp : _ ; ConfigLoad.scale = _")
    ld_i8 = autofission(ld_i8, ld_i8.find("ConfigLoad.scale = _").after(), n_lifts=3)
    ld_i8 = write_config(
        ld_i8,
        ld_i8.find("ConfigLoad.scale = _").after(),
        ConfigLoad,
        "src_stride",
        "stride(src, 0)",
    )
    ld_i8 = replace(ld_i8, "for i in _:_", do_ld_i8)
    ld_i8 = replace(
        ld_i8, "ConfigLoad.scale = _ ; ConfigLoad.src_stride = _", config_ld_i8
    )

    assert f"{config_ld_i8}\n{ld_i8}" == golden


"""

    @proc
    def ld_i8_v1_5(
        n     : size,
        m     : size,
        scale : f32,
        src   : [i8][n, m] @ DRAM,
        dst   : [i8][n, 16] @ GEMM_SCRATCH,
    ):
        assert n <= 16
        assert m <= 16
        assert stride(src, 1) == 1
        assert stride(dst, 0) == 16
        assert stride(dst, 1) == 1
        assert stride(src, 0) == ConfigLoad.src_stride

        ConfigLoad.scale = scale
        ConfigLoad.src_stride = stride(src, 0)

        for i in seq(0, n):
            for j in seq(0, m):
                tmp : f32
                tmp      = src[i,j]
                tmp      = tmp * ConfigLoad.scale
                dst[i,j] = tmp

    @proc
    def ld_i8_v2(
        n     : size,
        m     : size,
        scale : f32,
        src   : [i8][n, m] @ DRAM,
        dst   : [i8][n, 16] @ GEMM_SCRATCH,
    ):
        assert n <= 16
        assert m <= 16
        assert stride(src, 1) == 1
        assert stride(dst, 0) == 16
        assert stride(dst, 1) == 1
        assert stride(src, 0) == ConfigLoad.src_stride

        config_ld_i8(scale, stride(src, 0))
        do_ld_i8(n, m, src, dst)


    new = (ld_i8.rename("new")
                .replace(ld_i8, "for i in _:_")
                .call_eqv(ld_i8_v2, "ld_i8(_, _, _, _, _)"))
    print(new)




def test_config_write3():
    ConfigAB = new_config_f32()
    @proc
    def foo(n : size):
        for i in seq(0, n):
            ConfigAB.a = 3.0
            ConfigAB.b = ConfigAB.a



# This is fine
def test_read_write2():
    @proc
    def foo(n : size, A : i8[n]):
        a : i8
        a = 4.0
        for i in seq(0, n):
            a    = 0.0
            A[i] = a

    foo.check_effects()



def new_CONFIG_2f32():



@proc
def test_make_config(n: size, dst: R[n] @ DRAM, src: R[n] @ DRAM):
    for i in seq(0, (n + 7) / 8):
        if n - 8 * i >= 8:
            pass
        else:
            for j in seq(0, n - 8 * i):
                dst[8 * i + j] = src[8 * i + j]



@instr("write_config({a}, {b})")
def set_real_config( a : f32, b : f32):
    REAL_CONFIG.a = a
    REAL_CONFIG.b = b

# st_config_ex
@proc
def set_config_a( a : f32 ):
    DUMMY_CONFIG.a = a
    set_real_config(DUMMY_CONFIG.a, DUMMY_CONFIG.b )

# matmul_config_ex
@proc
def set_config_b( b : f32 ):
    DUMMY_CONFIG.b = b
    set_real_config(DUMMY_CONFIG.a, DUMMY_CONFIG.b )

@instr("mvin...")
def st_i8_v2( ..., stride : stride, ...)
    set_config_a(stride)
    do_st_i8(...)
"""



"""
This module defines special symbols that are used inside Exo code. You may
import this module via `from exo.syntax import *` to suppress warnings and
see documentation inside an IDE (like PyCharm).
"""

from typing import TypeVar as _TypeVar


def _fail_dummy():
    raise SyntaxError("Exo Python dummy functions should never be called " "directly.")


# noinspection PyPep8Naming
class size(int):
    """
    A type representing a size in Exo.
    """

    def __init__(self, *_):
        raise SyntaxError("Exo Python dummy objects should never be " "instantiated")


# noinspection PyPep8Naming
class f32(float):
    """
    A type representing a 32-bit floating point value.
    """

    def __init__(self, *_):
        raise SyntaxError("Exo Python dummy objects should never be " "instantiated")


def stride(tensor, dimension):
    """
    Given a tensor and a dimension, returns an opaque value representing the
    stride on that dimension.
    """
    _fail_dummy()


def par(lo, hi):
    """
    Parallel range between lo (must be 0) and hi (exclusive). May only be
    used in a for-loop bound. Iterations proceed as if in random order.
    """


# Common tensor names:
# A = _TypeVar('A', bound=size)
# B = _TypeVar('B', bound=size)
# C = _TypeVar('C', bound=size)
D = _TypeVar("D", bound=size)
E = _TypeVar("E", bound=size)
F = _TypeVar("F", bound=size)
G = _TypeVar("G", bound=size)
H = _TypeVar("H", bound=size)
I = _TypeVar("I", bound=size)
J = _TypeVar("J", bound=size)
K = _TypeVar("K", bound=size)
L = _TypeVar("L", bound=size)
M = _TypeVar("M", bound=size)
N = _TypeVar("N", bound=size)
O = _TypeVar("O", bound=size)
P = _TypeVar("P", bound=size)
Q = _TypeVar("Q", bound=size)
R = _TypeVar("R", bound=size)
S = _TypeVar("S", bound=size)
T = _TypeVar("T", bound=size)
U = _TypeVar("U", bound=size)
V = _TypeVar("V", bound=size)
W = _TypeVar("W", bound=size)
X = _TypeVar("X", bound=size)
Y = _TypeVar("Y", bound=size)
Z = _TypeVar("Z", bound=size)

a = _TypeVar("a", bound=size)
b = _TypeVar("b", bound=size)
c = _TypeVar("c", bound=size)
d = _TypeVar("d", bound=size)
e = _TypeVar("e", bound=size)
f = _TypeVar("f", bound=size)
g = _TypeVar("g", bound=size)
h = _TypeVar("h", bound=size)
# Common loop variable names:
# i = _TypeVar('i', bound=size)
# j = _TypeVar('j', bound=size)
# k = _TypeVar('k', bound=size)
l = _TypeVar("l", bound=size)
m = _TypeVar("m", bound=size)
n = _TypeVar("n", bound=size)
o = _TypeVar("o", bound=size)
p = _TypeVar("p", bound=size)
q = _TypeVar("q", bound=size)
r = _TypeVar("r", bound=size)
s = _TypeVar("s", bound=size)
t = _TypeVar("t", bound=size)
u = _TypeVar("u", bound=size)
v = _TypeVar("v", bound=size)
w = _TypeVar("w", bound=size)
x = _TypeVar("x", bound=size)
y = _TypeVar("y", bound=size)
z = _TypeVar("z", bound=size)



import functools
import re
import textwrap
from collections import ChainMap
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path

from .LoopIR import LoopIR, LoopIR_Do, get_writes_of_stmts, T, CIR
from .configs import ConfigError
from .mem_analysis import MemoryAnalysis
from .memory import MemGenError, Memory, DRAM, StaticMemory
from .parallel_analysis import ParallelAnalysis
from .prec_analysis import PrecisionAnalysis
from .prelude import *
from .win_analysis import WindowAnalysis
from .range_analysis import IndexRangeEnvironment


def sanitize_str(s):
    return re.sub(r"\W", "_", s)


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #

CacheDict = lambda: defaultdict(CacheDict)

op_prec = {
    "or": 10,
    #
    "and": 20,
    #
    "==": 30,
    #
    "<": 40,
    ">": 40,
    "<=": 40,
    ">=": 40,
    #
    "+": 50,
    "-": 50,
    #
    "*": 60,
    "/": 60,
    "%": 60,
    # unary minus
    "~": 70,
}


def lift_to_cir(e, range_env):
    assert e.type.is_indexable(), "why are you here?"

    is_non_neg = lambda e: range_env.check_expr_bound(0, IndexRangeEnvironment.leq, e)

    if isinstance(e, LoopIR.Read):
        return CIR.Read(e.name, is_non_neg(e))
    elif isinstance(e, LoopIR.Const):
        return CIR.Const(e.val)
    elif isinstance(e, LoopIR.BinOp):
        lhs = lift_to_cir(e.lhs, range_env)
        rhs = lift_to_cir(e.rhs, range_env)
        return CIR.BinOp(e.op, lhs, rhs, is_non_neg(e))
    elif isinstance(e, LoopIR.USub):
        arg = lift_to_cir(e.arg, range_env)
        return CIR.USub(arg, is_non_neg(e))
    else:
        assert False, "bad case!"


operations = {
    "+": lambda x, y: x + y,
    "-": lambda x, y: x - y,
    "*": lambda x, y: x * y,
    "/": lambda x, y: x / y,
    "%": lambda x, y: x % y,
}


def simplify_cir(e):
    if isinstance(e, (CIR.Read, CIR.Const, CIR.Stride)):
        return e

    elif isinstance(e, CIR.BinOp):
        lhs = simplify_cir(e.lhs)
        rhs = simplify_cir(e.rhs)

        if isinstance(lhs, CIR.Const) and isinstance(rhs, CIR.Const):
            return CIR.Const(operations[e.op](lhs.val, rhs.val))

        if isinstance(lhs, CIR.Const) and lhs.val == 0:
            if e.op == "+":
                return rhs
            elif e.op == "*" or e.op == "/":
                return CIR.Const(0)
            elif e.op == "-":
                pass  # cannot simplify
            else:
                assert False

        if isinstance(rhs, CIR.Const) and rhs.val == 0:
            if e.op == "+" or e.op == "-":
                return lhs
            elif e.op == "*":
                return CIR.Const(0)
            elif e.op == "/":
                assert False, "division by zero??"
            else:
                assert False, "bad case"

        if isinstance(lhs, CIR.Const) and lhs.val == 1 and e.op == "*":
            return rhs

        if isinstance(rhs, CIR.Const) and rhs.val == 1 and (e.op == "*" or e.op == "/"):
            return lhs

        return CIR.BinOp(e.op, lhs, rhs, e.is_non_neg)
    elif isinstance(e, CIR.USub):
        arg = simplify_cir(e.arg)
        if isinstance(arg, CIR.USub):
            return arg.arg
        if isinstance(arg, CIR.Const):
            return arg.update(val=-(arg.val))
        return e.update(arg=arg)
    else:
        assert False, "bad case!"


class LoopIR_SubProcs(LoopIR_Do):
    def __init__(self, proc):
        self._subprocs = set()
        if proc.instr is None:
            super().__init__(proc)

    def result(self):
        return self._subprocs

    # to improve efficiency
    def do_e(self, e):
        pass

    def do_s(self, s):
        if isinstance(s, LoopIR.Call):
            self._subprocs.add(s.f)
        else:
            super().do_s(s)


def find_all_subprocs(proc_list):
    all_procs = []
    seen = set()

    def walk(proc, visited):
        if proc in seen:
            return

        all_procs.append(proc)
        seen.add(proc)

        for sp in LoopIR_SubProcs(proc).result():
            if sp in visited:
                raise ValueError(f"found call cycle involving {sp.name}")
            walk(sp, visited | {proc})

    for proc in proc_list:
        walk(proc, set())

    # Reverse for C declaration order.
    return list(reversed(all_procs))


class LoopIR_FindMems(LoopIR_Do):
    def __init__(self, proc):
        self._mems = set()
        for a in proc.args:
            if a.mem:
                self._mems.add(a.mem)
        super().__init__(proc)

    def result(self):
        return self._mems

    # to improve efficiency
    def do_e(self, e):
        pass

    def do_s(self, s):
        if isinstance(s, LoopIR.Alloc):
            if s.mem:
                self._mems.add(s.mem)
        else:
            super().do_s(s)

    def do_t(self, t):
        pass


class LoopIR_FindBuiltIns(LoopIR_Do):
    def __init__(self, proc):
        self._builtins = set()
        super().__init__(proc)

    def result(self):
        return self._builtins

    # to improve efficiency
    def do_e(self, e):
        if isinstance(e, LoopIR.BuiltIn):
            self._builtins.add(e.f)
        else:
            super().do_e(e)

    def do_t(self, t):
        pass


class LoopIR_FindConfigs(LoopIR_Do):
    def __init__(self, proc):
        self._configs = set()
        super().__init__(proc)

    def result(self):
        return self._configs

    # to improve efficiency
    def do_e(self, e):
        if isinstance(e, LoopIR.ReadConfig):
            self._configs.add(e.config)
        else:
            super().do_e(e)

    def do_s(self, s):
        if isinstance(s, LoopIR.WriteConfig):
            self._configs.add(s.config)
        super().do_s(s)

    def do_t(self, t):
        pass


def find_all_mems(proc_list):
    mems = set()
    for p in proc_list:
        mems.update(LoopIR_FindMems(p).result())

    return [m for m in mems]


def find_all_builtins(proc_list):
    builtins = set()
    for p in proc_list:
        builtins.update(LoopIR_FindBuiltIns(p).result())

    return [b for b in builtins]


def find_all_configs(proc_list):
    configs = set()
    for p in proc_list:
        configs.update(LoopIR_FindConfigs(p).result())

    return list(configs)


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #


@dataclass(frozen=True)
class WindowStruct:
    name: str
    definition: str


@functools.cache
def _window_struct(typename, ctype, n_dims, is_const) -> WindowStruct:
    const_kwd = "const " if is_const else ""
    const_suffix = "c" if is_const else ""

    sname = f"exo_win_{n_dims}{typename}{const_suffix}"
    sdef = (
        f"struct {sname}{{\n"
        f"    {const_kwd}{ctype} * const data;\n"
        f"    const int_fast32_t strides[{n_dims}];\n"
        f"}};"
    )

    sdef_guard = sname.upper()
    sdef = f"""#ifndef {sdef_guard}
#define {sdef_guard}
{sdef}
#endif"""

    return WindowStruct(sname, sdef)


def window_struct(base_type, n_dims, is_const) -> WindowStruct:
    assert n_dims >= 1

    _window_struct_shorthand = {
        T.f16: "f16",
        T.f32: "f32",
        T.f64: "f64",
        T.i8: "i8",
        T.ui8: "ui8",
        T.ui16: "ui16",
        T.i32: "i32",
    }

    return _window_struct(
        _window_struct_shorthand[base_type], base_type.ctype(), n_dims, is_const
    )


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Loop IR Compiler Entry-points

# top level compiler function called by tests!


def run_compile(proc_list, h_file_name: str):
    file_stem = str(Path(h_file_name).stem)
    lib_name = sanitize_str(file_stem)
    fwd_decls, body = compile_to_strings(lib_name, proc_list)

    source = f'#include "{h_file_name}"\n\n{body}'

    header_guard = f"{lib_name}_H".upper()
    header = f"""
#pragma once
#ifndef {header_guard}
#define {header_guard}

#ifdef __cplusplus
extern "C" {{
#endif

{fwd_decls}

#ifdef __cplusplus
}}
#endif
#endif  // {header_guard}
"""

    return source, header


_static_helpers = {
    "exo_floor_div": textwrap.dedent(
        """
        static int exo_floor_div(int num, int quot) {
          int off = (num>=0)? 0 : quot-1;
          return (num-off)/quot;
        }
        """
    ),
}


def compile_to_strings(lib_name, proc_list):
    # Get transitive closure of call-graph
    orig_procs = [id(p) for p in proc_list]

    def from_lines(x):
        return "\n".join(x)

    proc_list = list(sorted(find_all_subprocs(proc_list), key=lambda x: x.name))

    # Header contents
    ctxt_name, ctxt_def = _compile_context_struct(find_all_configs(proc_list), lib_name)
    struct_defns = set()
    public_fwd_decls = []

    # Body contents
    memory_code = _compile_memories(find_all_mems(proc_list))
    builtin_code = _compile_builtins(find_all_builtins(proc_list))
    private_fwd_decls = []
    proc_bodies = []
    instrs_global = []

    needed_helpers = set()

    # Compile proc bodies
    seen_procs = set()
    for p in proc_list:
        if p.name in seen_procs:
            raise TypeError(f"multiple procs named {p.name}")
        seen_procs.add(p.name)

        # don't compile instruction procedures, but add a comment.
        if p.instr is not None:
            argstr = ",".join([str(a.name) for a in p.args])
            proc_bodies.extend(
                [
                    "",
                    '/* relying on the following instruction..."',
                    f"{p.name}({argstr})",
                    p.instr.c_instr,
                    "*/",
                ]
            )
            if p.instr.c_global:
                instrs_global.append(p.instr.c_global)
        else:
            is_public_decl = id(p) in orig_procs

            p = ParallelAnalysis().run(p)
            p = PrecisionAnalysis().run(p)
            p = WindowAnalysis().apply_proc(p)
            p = MemoryAnalysis().run(p)

            comp = Compiler(p, ctxt_name, is_public_decl=is_public_decl)
            d, b = comp.comp_top()
            struct_defns |= comp.struct_defns()
            needed_helpers |= comp.needed_helpers()

            if is_public_decl:
                public_fwd_decls.append(d)
            else:
                private_fwd_decls.append(d)

            proc_bodies.append(b)

    # Structs are just blobs of code... still sort them for output stability
    struct_defns = [x.definition for x in sorted(struct_defns, key=lambda x: x.name)]

    header_contents = f"""
#include <stdint.h>
#include <stdbool.h>

// Compiler feature macros adapted from Hedley (public domain)
// https://github.com/nemequ/hedley

#if defined(__has_builtin)
#  define EXO_HAS_BUILTIN(builtin) __has_builtin(builtin)
#else
#  define EXO_HAS_BUILTIN(builtin) (0)
#endif

#if EXO_HAS_BUILTIN(__builtin_assume)
#  define EXO_ASSUME(expr) __builtin_assume(expr)
#elif EXO_HAS_BUILTIN(__builtin_unreachable)
#  define EXO_ASSUME(expr) \\
      ((void)((expr) ? 1 : (__builtin_unreachable(), 1)))
#else
#  define EXO_ASSUME(expr) ((void)(expr))
#endif

{from_lines(ctxt_def)}
{from_lines(struct_defns)}
{from_lines(public_fwd_decls)}
"""

    helper_code = [_static_helpers[v] for v in needed_helpers]
    body_contents = [
        helper_code,
        instrs_global,
        memory_code,
        builtin_code,
        private_fwd_decls,
        proc_bodies,
    ]
    body_contents = list(filter(lambda x: x, body_contents))  # filter empty lines
    body_contents = map(from_lines, body_contents)
    body_contents = from_lines(body_contents)
    body_contents += "\n"  # New line at end of file
    return header_contents, body_contents


def _compile_builtins(builtins):
    builtin_code = []
    for b in sorted(builtins, key=lambda x: x.name()):
        if glb := b.globl():
            builtin_code.append(glb)
    return builtin_code


def _compile_memories(mems):
    memory_code = []
    for m in sorted(mems, key=lambda x: x.name()):
        memory_code.append(m.global_())
    return memory_code


def _compile_context_struct(configs, lib_name):
    if not configs:
        return "void", []

    ctxt_name = f"{lib_name}_Context"
    ctxt_def = [f"typedef struct {ctxt_name} {{ ", f""]

    seen = set()
    for c in sorted(configs, key=lambda x: x.name()):
        name = c.name()

        if name in seen:
            raise TypeError(f"multiple configs named {name}")
        seen.add(name)

        if c.is_allow_rw():
            sdef_lines = c.c_struct_def()
            sdef_lines = [f"    {line}" for line in sdef_lines]
            ctxt_def += sdef_lines
            ctxt_def += [""]
        else:
            ctxt_def += [f"// config '{name}' not materialized", ""]

    ctxt_def += [f"}} {ctxt_name};"]
    return ctxt_name, ctxt_def


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Loop IR Compiler


class Compiler:
    def __init__(self, proc, ctxt_name, *, is_public_decl):
        assert isinstance(proc, LoopIR.proc)

        self.proc = proc
        self.ctxt_name = ctxt_name
        self.env = ChainMap()
        self.range_env = IndexRangeEnvironment(proc, fast=False)
        self.names = ChainMap()
        self.envtyp = dict()
        self.mems = dict()
        self._tab = ""
        self._lines = []
        self._scalar_refs = set()
        self._needed_helpers = set()
        self.window_defns = set()
        self._known_strides = {}

        assert self.proc.name is not None, "expected names for compilation"
        name = self.proc.name
        arg_strs = []
        typ_comments = []

        # reserve the first "ctxt" argument
        self.new_varname(Sym("ctxt"), None)
        arg_strs.append(f"{ctxt_name} *ctxt")

        self.non_const = set(e for e, _ in get_writes_of_stmts(self.proc.body))

        for a in proc.args:
            mem = a.mem if a.type.is_numeric() else None
            name_arg = self.new_varname(a.name, typ=a.type, mem=mem)
            if a.type in (T.size, T.index, T.bool, T.stride):
                arg_strs.append(f"{a.type.ctype()} {name_arg}")
                typ_comments.append(f"{name_arg} : {a.type}")
            # setup, arguments
            else:
                assert a.type.is_numeric()
                assert a.type.basetype() != T.R
                if a.type.is_real_scalar():
                    self._scalar_refs.add(a.name)
                if a.type.is_win():
                    wintyp = self.get_window_type(a)
                    arg_strs.append(f"struct {wintyp} {name_arg}")
                else:
                    const_kwd = "const " if a.name not in self.non_const else ""
                    ctyp = a.type.basetype().ctype()
                    arg_strs.append(f"{const_kwd}{ctyp}* {name_arg}")
                mem = f" @{a.mem.name()}" if a.mem else ""
                comment_str = f"{name_arg} : {a.type}{mem}"
                typ_comments.append(comment_str)

        for pred in proc.preds:
            if isinstance(pred, LoopIR.Const):
                # TODO: filter these out earlier?
                continue

            if (
                isinstance(pred, LoopIR.BinOp)
                and pred.op == "=="
                and isinstance(pred.lhs, LoopIR.StrideExpr)
                and isinstance(pred.rhs, LoopIR.Const)
            ):
                self._known_strides[(pred.lhs.name, pred.lhs.dim)] = CIR.Const(
                    pred.rhs.val
                )
                self.add_line(f"// assert {pred}")
            else:
                # Default to just informing the compiler about the constraint
                # on a best-effort basis
                self.add_line(f"EXO_ASSUME({self.comp_e(pred)});")

        if not self.static_memory_check(self.proc):
            raise MemGenError("Cannot generate static memory in non-leaf procs")

        self.comp_stmts(self.proc.body)

        static_kwd = "" if is_public_decl else "static "

        # Generate headers here?
        comment = (
            f"// {name}(\n" + ",\n".join(["//     " + s for s in typ_comments]) + "\n"
            "// )\n"
        )
        proc_decl = comment + f"{static_kwd}void {name}( {', '.join(arg_strs)} );\n"
        proc_def = (
            comment
            + f"{static_kwd}void {name}( {', '.join(arg_strs)} ) {{\n"
            + "\n".join(self._lines)
            + "\n"
            "}\n"
        )

        self.proc_decl = proc_decl
        self.proc_def = proc_def

    def static_memory_check(self, proc):
        def allocates_static_memory(stmts):
            check = False
            for s in stmts:
                if isinstance(s, LoopIR.Alloc):
                    mem = s.mem
                    assert issubclass(mem, Memory)
                    check |= issubclass(mem, StaticMemory)
                elif isinstance(s, LoopIR.For):
                    check |= allocates_static_memory(s.body)
                elif isinstance(s, LoopIR.If):
                    check |= allocates_static_memory(s.body)
                    check |= allocates_static_memory(s.orelse)
            return check

        def is_leaf_proc(stmts):
            check = True
            for s in stmts:
                if isinstance(s, LoopIR.Call):
                    # Since intrinsics don't allocate memory, we can ignore
                    # them for leaf-node classification purposes. We want
                    # to avoid nested procs that both allocate static memory.
                    check &= s.f.instr is not None
                elif isinstance(s, LoopIR.For):
                    check &= is_leaf_proc(s.body)
                elif isinstance(s, LoopIR.If):
                    check &= is_leaf_proc(s.body)
                    check &= is_leaf_proc(s.orelse)
            return check

        return not allocates_static_memory(proc.body) or is_leaf_proc(proc.body)

    def add_line(self, line):
        if line:
            self._lines.append(self._tab + line)

    def comp_stmts(self, stmts):
        for b in stmts:
            self.comp_s(b)

    def comp_top(self):
        return self.proc_decl, self.proc_def

    def struct_defns(self):
        return self.window_defns

    def needed_helpers(self):
        return self._needed_helpers

    def new_varname(self, symbol, typ, mem=None):
        strnm = str(symbol)
        if strnm not in self.names:
            pass
        else:
            s = self.names[strnm]
            while s in self.names:
                m = re.match(r"^(.*)_([0-9]*)$", s)
                if not m:
                    s = s + "_1"
                else:
                    s = f"{m[1]}_{int(m[2]) + 1}"
            self.names[strnm] = s
            strnm = s

        self.names[strnm] = strnm
        self.env[symbol] = strnm
        self.envtyp[symbol] = typ
        if mem is not None:
            self.mems[symbol] = mem
        else:
            self.mems[symbol] = DRAM
        return strnm

    def push(self, only=None):
        if only is None:
            self.env = self.env.new_child()
            self.range_env.enter_scope()
            self.names = self.names.new_child()
            self._tab = self._tab + "  "
        elif only == "env":
            self.env = self.env.new_child()
            self.range_env.enter_scope()
            self.names = self.names.new_child()
        elif only == "tab":
            self._tab = self._tab + "  "
        else:
            assert False, f"BAD only parameter {only}"

    def pop(self):
        self.env = self.env.parents
        self.range_env.exit_scope()
        self.names = self.names.parents
        self._tab = self._tab[:-2]

    def comp_cir(self, e, env, prec) -> str:
        if isinstance(e, CIR.Read):
            return env[e.name]

        elif isinstance(e, CIR.Const):
            return str(e.val)

        elif isinstance(e, CIR.BinOp):
            local_prec = op_prec[e.op]

            lhs = self.comp_cir(e.lhs, env, local_prec)
            rhs = self.comp_cir(e.rhs, env, local_prec)

            if isinstance(e.rhs, CIR.BinOp) and (e.op == "-" or e.op == "/"):
                rhs = f"({rhs})"

            if e.op == "/":
                if (isinstance(e.lhs, (CIR.Read, CIR.BinOp)) and e.lhs.is_non_neg) or (
                    isinstance(e.lhs, CIR.Const) and e.lhs.val > 0
                ):
                    return f"({lhs} / {rhs})"
                else:
                    return self._call_static_helper("exo_floor_div", lhs, rhs)

            s = f"{lhs} {e.op} {rhs}"
            if local_prec < prec:
                s = f"({s})"

            return s

        elif isinstance(e, CIR.Stride):
            return f"{e.name}.strides[{e.dim}]"
        elif isinstance(e, CIR.USub):
            return f'-{self.comp_cir(e.arg, env, op_prec["~"])}'
        else:
            assert False, "bad case!"

    def access_str(self, nm, idx_list) -> str:
        type = self.envtyp[nm]
        cirs = [lift_to_cir(i, self.range_env) for i in idx_list]
        idx_expr = self.get_idx_offset(nm, type, cirs)
        idx_expr_s = self.comp_cir(simplify_cir(idx_expr), self.env, prec=0)
        buf = self.env[nm]
        if not type.is_win():
            return f"{buf}[{idx_expr_s}]"
        else:
            return f"{buf}.data[{idx_expr_s}]"

    def shape_strs(self, shape, prec=100) -> str:
        comp_res = [
            self.comp_cir(simplify_cir(lift_to_cir(i, self.range_env)), self.env, prec)
            for i in shape
        ]
        return comp_res

    def tensor_strides(self, shape) -> CIR:
        szs = [lift_to_cir(i, self.range_env) for i in shape]
        assert len(szs) >= 1
        strides = [CIR.Const(1)]
        s = szs[-1]
        for sz in reversed(szs[:-1]):
            strides.append(s)
            s = CIR.BinOp("*", sz, s, True)
        strides = list(reversed(strides))

        return strides

    # works for any tensor or window type
    def get_strides(self, name: Sym, typ) -> CIR:
        if typ.is_win():
            res = []
            for i in range(len(typ.shape())):
                if stride := self._known_strides.get((name, i)):
                    res.append(stride)
                else:
                    res.append(CIR.Stride(name, i))

            return res
        else:
            return self.tensor_strides(typ.shape())

    def get_idx_offset(self, name: Sym, typ, idx) -> CIR:
        strides = self.get_strides(name, typ)
        assert len(strides) == len(idx)
        acc = CIR.BinOp("*", idx[0], strides[0], True)
        for i, s in zip(idx[1:], strides[1:]):
            new = CIR.BinOp("*", i, s, True)
            acc = CIR.BinOp("+", acc, new, True)

        return acc

    def get_window_type(self, typ, is_const=None):
        assert isinstance(typ, T.Window) or (
            isinstance(typ, LoopIR.fnarg) and typ.type.is_win()
        )

        if isinstance(typ, T.Window):
            base = typ.as_tensor.basetype()
            n_dims = len(typ.as_tensor.shape())
            if is_const is None:
                is_const = typ.src_buf not in self.non_const
        else:
            base = typ.type.basetype()
            n_dims = len(typ.type.shape())
            if is_const is None:
                is_const = typ.name not in self.non_const

        win = window_struct(base, n_dims, is_const)
        self.window_defns.add(win)
        return win.name

    def comp_s(self, s):
        if isinstance(s, LoopIR.Pass):
            self.add_line("; // NO-OP")
        elif isinstance(s, (LoopIR.Assign, LoopIR.Reduce)):
            if s.name in self._scalar_refs:
                lhs = f"*{self.env[s.name]}"
            elif self.envtyp[s.name].is_real_scalar():
                lhs = self.env[s.name]
            else:
                lhs = self.access_str(s.name, s.idx)
            rhs = self.comp_e(s.rhs)

            # possibly cast!
            lbtyp = s.type.basetype()
            rbtyp = s.rhs.type.basetype()
            if lbtyp != rbtyp:
                assert s.type.is_real_scalar()
                assert s.rhs.type.is_real_scalar()

                rhs = f"({lbtyp.ctype()})({rhs})"

            mem: Memory = self.mems[s.name]
            if isinstance(s, LoopIR.Assign):
                self.add_line(mem.write(s, lhs, rhs))
            else:
                self.add_line(mem.reduce(s, lhs, rhs))

        elif isinstance(s, LoopIR.WriteConfig):
            if not s.config.is_allow_rw():
                raise ConfigError(
                    f"{s.srcinfo}: cannot write to config '{s.config.name()}'"
                )

            nm = s.config.name()
            rhs = self.comp_e(s.rhs)

            # possibly cast!
            ltyp = s.config.lookup(s.field)[1]
            rtyp = s.rhs.type
            if ltyp != rtyp and not ltyp.is_indexable():
                assert ltyp.is_real_scalar()
                assert rtyp.is_real_scalar()

                rhs = f"({ltyp.ctype()})({rhs})"

            self.add_line(f"ctxt->{nm}.{s.field} = {rhs};")

        elif isinstance(s, LoopIR.WindowStmt):
            win_struct = self.get_window_type(s.rhs.type)
            rhs = self.comp_e(s.rhs)
            assert isinstance(s.rhs, LoopIR.WindowExpr)
            mem = self.mems[s.rhs.name]
            name = self.new_varname(s.name, typ=s.rhs.type, mem=mem)
            self.add_line(f"struct {win_struct} {name} = {rhs};")
        elif isinstance(s, LoopIR.If):
            cond = self.comp_e(s.cond)
            self.add_line(f"if ({cond}) {{")
            self.push()
            self.comp_stmts(s.body)
            self.pop()
            if len(s.orelse) > 0:
                self.add_line("} else {")
                self.push()
                self.comp_stmts(s.orelse)
                self.pop()
            self.add_line("}")

        elif isinstance(s, LoopIR.For):
            lo = self.comp_e(s.lo)
            hi = self.comp_e(s.hi)
            self.push(only="env")
            itr = self.new_varname(s.iter, typ=T.index)  # allocate a new string
            self.range_env.add_loop_iter(
                s.iter,
                s.lo,
                s.hi,
            )
            if isinstance(s.loop_mode, LoopIR.Par):
                self.add_line(f"#pragma omp parallel for")
            self.add_line(f"for (int_fast32_t {itr} = {lo}; {itr} < {hi}; {itr}++) {{")
            self.push(only="tab")
            self.comp_stmts(s.body)
            self.pop()
            self.add_line("}")

        elif isinstance(s, LoopIR.Alloc):
            name = self.new_varname(s.name, typ=s.type, mem=s.mem)
            assert s.type.basetype().is_real_scalar()
            assert s.type.basetype() != T.R
            ctype = s.type.basetype().ctype()
            mem = s.mem or DRAM
            line = mem.alloc(name, ctype, self.shape_strs(s.type.shape()), s.srcinfo)

            self.add_line(line)
        elif isinstance(s, LoopIR.Free):
            name = self.env[s.name]
            assert s.type.basetype().is_real_scalar()
            ctype = s.type.basetype().ctype()
            mem = s.mem or DRAM
            line = mem.free(name, ctype, self.shape_strs(s.type.shape()), s.srcinfo)
            self.add_line(line)
        elif isinstance(s, LoopIR.Call):
            assert all(
                a.type.is_win() == fna.type.is_win() for a, fna in zip(s.args, s.f.args)
            )
            args = [self.comp_fnarg(e, s.f, i) for i, e in enumerate(s.args)]
            if s.f.instr is not None:
                d = dict()
                assert len(s.f.args) == len(args)
                for i in range(len(args)):
                    arg_name = str(s.f.args[i].name)
                    d[arg_name] = f"({args[i]})"
                    arg_type = s.args[i].type
                    if arg_type.is_win():
                        assert isinstance(s.args[i], LoopIR.WindowExpr)
                        data, _ = self.window_struct_fields(s.args[i])
                        d[f"{arg_name}_data"] = data
                        d[f"{arg_name}_int"] = self.env[s.args[i].name]
                    else:
                        d[f"{arg_name}_data"] = f"({args[i]})"

                self.add_line(f"{s.f.instr.c_instr.format(**d)}")
            else:
                fname = s.f.name
                args = ["ctxt"] + args
                self.add_line(f"{fname}({','.join(args)});")
        else:
            assert False, "bad case"

    def comp_fnarg(self, e, fn, i, *, prec=0):
        if isinstance(e, LoopIR.Read):
            assert not e.idx
            rtyp = self.envtyp[e.name]
            if rtyp.is_indexable():
                return self.env[e.name]
            elif rtyp is T.bool:
                return self.env[e.name]
            elif rtyp is T.stride:
                return self.env[e.name]
            elif e.name in self._scalar_refs:
                return self.env[e.name]
            elif rtyp.is_tensor_or_window():
                return self.env[e.name]
            else:
                assert rtyp.is_real_scalar()
                return f"&{self.env[e.name]}"
        elif isinstance(e, LoopIR.WindowExpr):
            if isinstance(fn, LoopIR.proc):
                callee_buf = fn.args[i].name
                is_const = callee_buf not in set(
                    x for x, _ in get_writes_of_stmts(fn.body)
                )
            else:
                raise NotImplementedError("Passing windows to built-ins")
            win_struct = self.get_window_type(e.type, is_const)
            data, strides = self.window_struct_fields(e)
            return f"(struct {win_struct}){{ &{data}, {{ {strides} }} }}"
        else:
            return self.comp_e(e, prec)

    def comp_e(self, e, prec=0):
        if isinstance(e, LoopIR.Read):
            rtyp = self.envtyp[e.name]
            if rtyp.is_indexable() or rtyp is T.bool or rtyp == T.stride:
                return self.env[e.name]

            mem: Memory = self.mems[e.name]

            if not mem.can_read():
                raise MemGenError(
                    f"{e.srcinfo}: cannot read from buffer "
                    f"'{e.name}' in memory '{mem.name()}'"
                )

            if e.name in self._scalar_refs:
                return f"*{self.env[e.name]}"
            elif not rtyp.is_tensor_or_window():
                return self.env[e.name]
            else:
                return self.access_str(e.name, e.idx)

        elif isinstance(e, LoopIR.WindowExpr):
            win_struct = self.get_window_type(e.type)
            data, strides = self.window_struct_fields(e)
            return f"(struct {win_struct}){{ &{data}, {{ {strides} }} }}"

        elif isinstance(e, LoopIR.Const):
            if isinstance(e.val, bool):
                return "true" if e.val else "false"
            elif e.type.is_indexable():
                return f"{int(e.val)}"
            elif e.type == T.f64:
                return f"{float(e.val)}"
            elif e.type == T.f32:
                return f"{float(e.val)}f"
            else:
                return f"(({e.type.ctype()}) {str(e.val)})"

        elif isinstance(e, LoopIR.BinOp):
            local_prec = op_prec[e.op]
            int_div = e.op == "/" and not e.type.is_numeric()
            if int_div:
                local_prec = 0
            op = e.op
            if op == "and":
                op = "&&"
            elif op == "or":
                op = "||"

            lhs = self.comp_e(e.lhs, local_prec)
            rhs = self.comp_e(e.rhs, local_prec + 1)

            if int_div:
                if self.range_env.check_expr_bound(0, IndexRangeEnvironment.leq, e):
                    # TODO: too many parens?
                    return f"(({lhs}) / ({rhs}))"
                return self._call_static_helper("exo_floor_div", lhs, rhs)

            s = f"{lhs} {op} {rhs}"
            if local_prec < prec:
                s = f"({s})"

            return s
        elif isinstance(e, LoopIR.USub):
            return f'-{self.comp_e(e.arg, op_prec["~"])}'

        elif isinstance(e, LoopIR.BuiltIn):
            args = [self.comp_fnarg(a, e, i) for i, a in enumerate(e.args)]
            return e.f.compile(args)

        elif isinstance(e, LoopIR.StrideExpr):
            basetyp = self.envtyp[e.name]
            stride = self.get_strides(e.name, basetyp)[e.dim]
            return self.comp_cir(simplify_cir(stride), self.env, prec=0)

        elif isinstance(e, LoopIR.ReadConfig):
            if not e.config.is_allow_rw():
                raise ConfigError(
                    f"{e.srcinfo}: cannot read from config '{e.config.name()}'"
                )
            return f"ctxt->{e.config.name()}.{e.field}"

        else:
            assert False, "bad case"

    def _call_static_helper(self, helper, *args):
        self._needed_helpers.add(helper)
        return f'{helper}({", ".join(map(str, args))})'

    def window_struct_fields(self, e):
        base = self.env[e.name]
        basetyp = self.envtyp[e.name]
        mem: Memory = self.mems[e.name]

        # compute offset to new data pointer
        def w_lo(w):
            return w.lo if isinstance(w, LoopIR.Interval) else w.pt

        cirs = [lift_to_cir(w_lo(w), self.range_env) for w in e.idx]
        idxs = [self.comp_cir(simplify_cir(i), self.env, prec=0) for i in cirs]

        # compute new window strides
        all_strides = self.get_strides(e.name, basetyp)
        all_strides_s = [
            self.comp_cir(simplify_cir(i), self.env, prec=0) for i in all_strides
        ]
        assert 0 < len(all_strides_s) == len(e.idx)
        dataptr = mem.window(basetyp, base, idxs, all_strides_s, e.srcinfo)
        strides = ", ".join(
            s for s, w in zip(all_strides_s, e.idx) if isinstance(w, LoopIR.Interval)
        )
        return dataptr, strides



from __future__ import annotations

import pytest

from exo.new_eff import *

from exo import proc, config, DRAM, SchedulingError
from exo.stdlib.scheduling import *
from exo.platforms.x86 import *


print()
print("Dev Tests for new_eff.py")


def test_debug_let_and_mod():
    N = AInt(Sym("N"))
    j = AInt(Sym("j"))
    i = AInt(Sym("i"))
    x = AInt(Sym("x"))

    F = A.ForAll(
        i.name,
        A.Let(
            [x.name],
            [A.Let([j.name], [AInt(64) * i], N + j, T.index, j.srcinfo)],
            AEq(x % AInt(64), AInt(0)),
            T.bool,
            x.srcinfo,
        ),
        T.bool,
        i.srcinfo,
    )

    slv = SMTSolver(verbose=True)

    slv.verify(F)


def test_reorder_stmts_fail():
    @proc
    def foo(N: size, x: R[N]):
        x[0] = 3.0
        x[0] = 4.0

    with pytest.raises(SchedulingError, match="do not commute"):
        foo = reorder_stmts(foo, "x[0] = 3.0 ; x[0] = 4.0")


def test_reorder_alloc_fail():
    @proc
    def foo(N: size, x: R[N]):
        y: R
        y = 4.0
        x[0] = y

    with pytest.raises(SchedulingError, match="do not commute"):
        foo = reorder_stmts(foo, "y : R ; y = 4.0")


def test_reorder_loops_success(golden):
    @proc
    def foo(N: size, x: R[N, N]):
        for i in seq(0, N):
            for j in seq(0, N):
                x[i, j] = x[i, j] * 2.0

    foo = reorder_loops(foo, "i j")
    assert str(foo) == golden


def test_reorder_loops_fail():
    @proc
    def foo(N: size, x: R[N, N]):
        for i in seq(0, N):
            for j in seq(0, N):
                x[i, j] = x[j, i] * 2.0

    with pytest.raises(SchedulingError, match="cannot be reordered"):
        foo = reorder_loops(foo, "i j")


def test_alloc_success(golden):
    @proc
    def foo(N: size, x: R[N, N]):
        for i in seq(0, N):
            for j in seq(0, N):
                tmp: R
                tmp = x[i, j] * 2.0
                x[i, j] = tmp

    foo = reorder_loops(foo, "i j")
    assert str(foo) == golden


def test_reorder_loops_requiring_seq(golden):
    # the stencil pattern here looks like
    #     o     o
    #       \   |
    #         \ V
    #     o --> x
    #
    # so it isn't safe to _reverse_
    # the iteration order, but it is safe
    # to reorder the loops

    @proc
    def foo(N: size, x: R[N, N]):
        for i in seq(0, N):
            for j in seq(0, N):
                if i > 0 and j > 0:
                    x[i, j] += (
                        -1.0 / 3.0 * (x[i - 1, j] + x[i - 1, j - 1] + x[i, j - 1])
                    )

    foo = reorder_loops(foo, "i j")
    assert str(foo) == golden


def test_reorder_loops_4pt_stencil_succeed(golden):
    # Also, a 4-point stencil being
    # used in e.g. a Gauss-Seidel scheme can be reordered

    @proc
    def foo(N: size, x: R[N, N]):
        for i in seq(0, N):
            for j in seq(0, N):
                if 0 < i < N - 1 and 0 < j < N - 1:
                    x[i, j] += (
                        -1.0
                        / 4.0
                        * (x[i - 1, j] + x[i + 1, j] + x[i, j - 1] + x[i, j + 1])
                    )

    foo = reorder_loops(foo, "i j")
    assert str(foo) == golden


def test_reorder_loops_failing_seq():
    # But if we do the stencil over the 4 diagonals, then it's not safe

    @proc
    def foo(N: size, x: R[N, N]):
        for i in seq(0, N):
            for j in seq(0, N):
                if 0 < i < N - 1 and 0 < j < N - 1:
                    x[i, j] += (
                        -1.0
                        / 4.0
                        * (
                            x[i - 1, j - 1]
                            + x[i - 1, j + 1]
                            + x[i + 1, j - 1]
                            + x[i + 1, j + 1]
                        )
                    )

    with pytest.raises(SchedulingError, match="cannot be reordered"):
        foo = reorder_loops(foo, "i j")


# Should add a test that shows that READing something in an assertion
# does in fact count as a READ effect for the procedure, but not
# for its body.  This can probably distinguish whether certain
# rewrites are allowed or not.


def test_delete_config_basic(golden):
    @config
    class CFG:
        a: index
        b: size

    @proc
    def foo(N: size, x: R[N]):
        CFG.a = 3
        for i in seq(0, N):
            x[i] = x[i] + 1.0

    foo = delete_config(foo, "CFG.a = _")
    assert str(foo) == golden


def test_delete_config_subproc_basic(golden):
    @config
    class CFG:
        a: index
        b: size

    @proc
    def do_config():
        CFG.a = 3
        CFG.b = 5

    @proc
    def foo(N: size, x: R[N]):
        do_config()
        for i in seq(0, N):
            x[i] = x[i] + 1.0

    foo = delete_config(foo, "do_config()")
    assert str(foo) == golden


def test_delete_config_fail():
    @config
    class CFG:
        a: index
        b: size

    @proc
    def foo(N: size, x: R[N]):
        CFG.a = 3
        for i in seq(0, N):
            if i < CFG.a:
                x[i] = x[i] + 1.0

    with pytest.raises(
        SchedulingError, match="Cannot change configuration value of CFG_a"
    ):
        foo = delete_config(foo, "CFG.a = _")


def test_delete_config_subproc_fail():
    @config
    class CFG:
        a: index
        b: size

    @proc
    def do_config():
        CFG.a = 3
        CFG.b = 5

    @proc
    def foo(N: size, x: R[N]):
        do_config()
        for i in seq(0, N):
            if i < CFG.a:
                x[i] = x[i] + 1.0

    with pytest.raises(
        SchedulingError, match="Cannot change configuration value of CFG_a"
    ):
        foo = delete_config(foo, "do_config()")


def test_delete_config_bc_shadow(golden):
    @config
    class CFG:
        a: index
        b: size

    @proc
    def foo(N: size, x: R[N]):
        CFG.a = 34
        CFG.a = 3
        for i in seq(0, N):
            if i < CFG.a:
                x[i] = x[i] + 1.0

    foo = delete_config(foo, "CFG.a = _ #0")
    assert str(foo) == golden


def test_delete_config_bc_redundant(golden):
    @config
    class CFG:
        a: index
        b: size

    @proc
    def foo(N: size, x: R[N]):
        CFG.a = 3
        CFG.a = 3
        for i in seq(0, N):
            if i < CFG.a:
                x[i] = x[i] + 1.0

    foo = delete_config(foo, "CFG.a = _ #1")
    assert str(foo) == golden


def test_delete_config_fail_bc_not_redundant():
    @config
    class CFG:
        a: index
        b: size

    @proc
    def foo(N: size, x: R[N]):
        CFG.a = 34
        CFG.a = 3
        for i in seq(0, N):
            if i < CFG.a:
                x[i] = x[i] + 1.0

    with pytest.raises(
        SchedulingError, match="Cannot change configuration value of CFG_a"
    ):
        foo = delete_config(foo, "CFG.a = _ #1")


def test_scal():
    @proc
    def sscal(n: size, alpha: f32 @ DRAM, x: [f32][n] @ DRAM):
        assert stride(x, 0) == 1
        reg0: f32[8] @ AVX2
        if n / 8 > 0:
            mm256_broadcast_ss_scalar(reg0[0:8], alpha)
        reg1: f32[8] @ AVX2
        reg2: f32[8] @ AVX2
        for io in seq(0, n / 8):
            mm256_loadu_ps(reg1[0:8], x[8 * io + 0 : 8 * io + 8])
            mm256_mul_ps(reg2[0:8], reg0[0:8], reg1[0:8])
            mm256_storeu_ps(x[8 * io + 0 : 8 * io + 8], reg2[0:8])
        for ii in seq(0, n % 8):
            x[ii + n / 8 * 8] = alpha * x[ii + n / 8 * 8]

    stmt_c = sscal.find("mm256_mul_ps(_)")

    with pytest.raises(SchedulingError):
        sscal = reorder_stmts(sscal, stmt_c.expand(1, 0))


def test_alias_check():
    """This fails when we have an aliasing in a proc"""

    @proc
    def foo(N: size, x: [f32][N], y: [f32][N]):
        for i in seq(0, N):
            x[i] = y[i]

    with pytest.raises(SchedulingError, match="Cannot Pass the same buffer"):

        @proc
        def bar(N: size, x: [f32][N]):
            foo(N, x, x)



from __future__ import annotations

from exo import proc, DRAM, QAST
from exo.stdlib.scheduling import *


# ------- Reflection tests ---------


def new_sgemm():
    @proc
    def sgemm_full(
        N: size,
        M: size,
        K: size,
        C: f32[N, M] @ DRAM,
        A: f32[N, K] @ DRAM,
        B: f32[K, M] @ DRAM,
    ):
        for i in seq(0, N):
            for j in seq(0, M):
                for k in seq(0, K):
                    C[i, j] += A[i, k] * B[k, j]

    return sgemm_full


def test_proc_name():
    sgemm = new_sgemm()

    proc = sgemm.get_ast()
    assert isinstance(proc, QAST.Proc)
    assert proc.name == "sgemm_full"

    sgemm = rename(sgemm, "sgemm")

    proc = sgemm.get_ast()
    assert isinstance(proc, QAST.Proc)
    assert proc.name == "sgemm"


def test_find_outer_loop():
    sgemm = new_sgemm()

    loops = sgemm.get_ast("for _ in _: _ #0")
    assert isinstance(loops, list) and len(loops) == 1
    assert isinstance(loops[0], QAST.For)

    assert loops[0].name == "i"
    i_body = loops[0].body
    assert isinstance(i_body, list) and len(i_body) == 1
    assert isinstance(i_body[0], QAST.For)

    assert i_body[0].name == "j"
    j_body = i_body[0].body
    assert isinstance(j_body, list) and len(j_body) == 1
    assert isinstance(j_body[0], QAST.For)

    assert j_body[0].name == "k"
    k_body = j_body[0].body
    assert isinstance(k_body, list) and len(k_body) == 1

    assert not isinstance(k_body[0], QAST.For)


def get_loop_nest_info(p, pattern):
    loops = p.get_ast(pattern)
    if loops is None:
        return []
    assert isinstance(loops, list) and len(loops) > 0
    assert isinstance(loops[0], QAST.Stmt), "must call with ... #_ pattern"

    def recurse_loops(loops):
        if len(loops) != 1:
            return []
        stmt = loops[0]
        assert isinstance(stmt, QAST.Stmt)

        if isinstance(stmt, QAST.For):
            return [(stmt.name, stmt.hi)] + recurse_loops(stmt.body)
        else:
            return []

    return recurse_loops(loops)


def test_get_outer_loop_info():
    sgemm = new_sgemm()

    info = get_loop_nest_info(sgemm, "for _ in _: _ #0")

    expect_info = [
        ("i", QAST.Read("N", [], QAST.size())),
        ("j", QAST.Read("M", [], QAST.size())),
        ("k", QAST.Read("K", [], QAST.size())),
    ]
    assert info == expect_info


def test_get_mid_loop_info():
    sgemm = new_sgemm()

    expect_info = [
        ("j", QAST.Read("M", [], QAST.size())),
        ("k", QAST.Read("K", [], QAST.size())),
    ]

    info = get_loop_nest_info(sgemm, "for j in _: _ #0")
    assert info == expect_info

    info = get_loop_nest_info(sgemm, "for _ in _: _ #1")
    assert info == expect_info


def test_get_bottom_loop_info():
    sgemm = new_sgemm()

    expect_info = [
        ("k", QAST.Read("K", [], QAST.size())),
    ]

    info = get_loop_nest_info(sgemm, "for k in _: _ #0")
    assert info == expect_info

    info = get_loop_nest_info(sgemm, "for _ in _: _ #2")
    assert info == expect_info


def test_get_no_loop_info():
    sgemm = new_sgemm()

    info = get_loop_nest_info(sgemm, "for abc in _: _ #0")
    assert info == []


def test_show_effect(golden):
    sgemm = new_sgemm()

    assert sgemm.show_effect("for j in _: _") == golden



from __future__ import annotations

from .. import instr, DRAM
from ..libs.memories import AVX2, AVX512


# --------------------------------------------------------------------------- #
#   AVX2 intrinsics
# --------------------------------------------------------------------------- #


@instr("{dst_data} = _mm256_setzero_ps();")
def mm256_setzero_ps(dst: [f32][8] @ AVX2):
    assert stride(dst, 0) == 1

    for i in seq(0, 8):
        dst[i] = 0.0


@instr("{dst_data} = _mm256_setzero_pd();")
def mm256_setzero_pd(dst: [f64][4] @ AVX2):
    assert stride(dst, 0) == 1

    for i in seq(0, 4):
        dst[i] = 0.0


@instr("{dst_data} = _mm256_loadu_ps(&{src_data});")
def mm256_loadu_ps(dst: [f32][8] @ AVX2, src: [f32][8] @ DRAM):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 8):
        dst[i] = src[i]


@instr("{dst_data} = _mm256_loadu_pd(&{src_data});")
def mm256_loadu_pd(dst: [f64][4] @ AVX2, src: [f64][4] @ DRAM):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 4):
        dst[i] = src[i]


@instr("_mm256_storeu_ps(&{dst_data}, {src_data});")
def mm256_storeu_ps(dst: [f32][8] @ DRAM, src: [f32][8] @ AVX2):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 8):
        dst[i] = src[i]


@instr("_mm256_storeu_pd(&{dst_data}, {src_data});")
def mm256_storeu_pd(dst: [f64][4] @ DRAM, src: [f64][4] @ AVX2):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 4):
        dst[i] = src[i]


@instr("{dst_data} = _mm256_fmadd_ps({src1_data}, {src2_data}, {dst_data});")
def mm256_fmadd_ps(
    dst: [f32][8] @ AVX2,
    src1: [f32][8] @ AVX2,
    src2: [f32][8] @ AVX2,
):
    assert stride(src1, 0) == 1
    assert stride(src2, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 8):
        dst[i] += src1[i] * src2[i]


@instr("{dst_data} = _mm256_fmadd_pd({src1_data}, {src2_data}, {dst_data});")
def mm256_fmadd_pd(
    dst: [f64][4] @ AVX2,
    src1: [f64][4] @ AVX2,
    src2: [f64][4] @ AVX2,
):
    assert stride(src1, 0) == 1
    assert stride(src2, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 4):
        dst[i] += src1[i] * src2[i]


@instr("{out_data} = _mm256_broadcast_ss(&{val_data});")
def mm256_broadcast_ss(
    out: [f32][8] @ AVX2,
    val: [f32][1],
):
    assert stride(out, 0) == 1

    for i in seq(0, 8):
        out[i] = val[0]


@instr("{out_data} = _mm256_broadcast_sd(&{val_data});")
def mm256_broadcast_sd(
    out: [f64][4] @ AVX2,
    val: [f64][1],
):
    assert stride(out, 0) == 1

    for i in seq(0, 4):
        out[i] = val[0]


@instr("{out_data} = _mm256_broadcast_ss({val_data});")
def mm256_broadcast_ss_scalar(out: [f32][8] @ AVX2, val: f32):
    assert stride(out, 0) == 1

    for i in seq(0, 8):
        out[i] = val


@instr("{out_data} = _mm256_broadcast_sd({val_data});")
def mm256_broadcast_sd_scalar(out: [f64][4] @ AVX2, val: f64):
    assert stride(out, 0) == 1

    for i in seq(0, 4):
        out[i] = val


@instr("{dst_data} = _mm512_fmadd_ps({dst_data}, {lhs_data}, {rhs_data});")
def mm256_fmadd_ps_broadcast(
    dst: [f32][8] @ AVX2, lhs: [f32][8] @ AVX2, rhs: [f32][1] @ DRAM
):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1

    for i in seq(0, 8):
        dst[i] += lhs[i] * rhs[0]


@instr("{out_data} = _mm256_mul_ps({x_data}, {y_data});")
def mm256_mul_ps(out: [f32][8] @ AVX2, x: [f32][8] @ AVX2, y: [f32][8] @ AVX2):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(y, 0) == 1

    for i in seq(0, 8):
        out[i] = x[i] * y[i]


@instr("{out_data} = _mm256_mul_pd({x_data}, {y_data});")
def mm256_mul_pd(out: [f64][4] @ AVX2, x: [f64][4] @ AVX2, y: [f64][4] @ AVX2):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(y, 0) == 1

    for i in seq(0, 4):
        out[i] = x[i] * y[i]


@instr("{out_data} = _mm256_div_ps({x_data}, {y_data});")
def mm256_div_ps(out: [f32][8] @ AVX2, x: [f32][8] @ AVX2, y: [f32][8] @ AVX2):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(y, 0) == 1

    for i in seq(0, 8):
        out[i] = x[i] / y[i]


@instr("{out_data} = _mm256_div_pd({x_data}, {y_data});")
def mm256_div_pd(out: [f64][4] @ AVX2, x: [f64][4] @ AVX2, y: [f64][4] @ AVX2):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(y, 0) == 1

    for i in seq(0, 4):
        out[i] = x[i] / y[i]


@instr("{out_data} = _mm256_add_ps({x_data}, {y_data});")
def mm256_add_ps(out: [f32][8] @ AVX2, x: [f32][8] @ AVX2, y: [f32][8] @ AVX2):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(y, 0) == 1

    for i in seq(0, 8):
        out[i] = x[i] + y[i]


@instr("{out_data} = _mm256_add_pd({x_data}, {y_data});")
def mm256_add_pd(out: [f64][4] @ AVX2, x: [f64][4] @ AVX2, y: [f64][4] @ AVX2):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(y, 0) == 1

    for i in seq(0, 4):
        out[i] = x[i] + y[i]


@instr("{out_data} = _mm256_sub_ps({x_data}, {y_data});")
def mm256_sub_ps(out: [f32][8] @ AVX2, x: [f32][8] @ AVX2, y: [f32][8] @ AVX2):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(y, 0) == 1

    for i in seq(0, 8):
        out[i] = x[i] - y[i]


@instr("{out_data} = _mm256_sub_pd({x_data}, {y_data});")
def mm256_sub_pd(out: [f64][4] @ AVX2, x: [f64][4] @ AVX2, y: [f64][4] @ AVX2):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(y, 0) == 1

    for i in seq(0, 4):
        out[i] = x[i] - y[i]


@instr("{dst_data} = _mm256_loadu_si256((const __m256i *) &{src_data});")
def mm256_loadu_si256(dst: [ui16][16] @ AVX2, src: [ui16][16] @ DRAM):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 16):
        dst[i] = src[i]


@instr("_mm256_storeu_si256((__m256i *) &{dst_data}, {src_data});")
def mm256_storeu_si256(dst: [ui16][16] @ DRAM, src: [ui16][16] @ AVX2):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 16):
        dst[i] = src[i]


@instr("{out_data} = _mm256_adds_epu16({x_data}, {y_data});")
def mm256_add_epi16(out: [ui16][16] @ AVX2, x: [ui16][16] @ AVX2, y: [ui16][16] @ AVX2):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(y, 0) == 1

    for i in seq(0, 16):
        out[i] = x[i] + y[i]


# --------------------------------------------------------------------------- #
#   AVX512 intrinsics
# --------------------------------------------------------------------------- #


@instr("{dst_data} = _mm512_loadu_ps(&{src_data});")
def mm512_loadu_ps(dst: [f32][16] @ AVX512, src: [f32][16] @ DRAM):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 16):
        dst[i] = src[i]


@instr("_mm512_storeu_ps(&{dst_data}, {src_data});")
def mm512_storeu_ps(dst: [f32][16] @ DRAM, src: [f32][16] @ AVX512):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 16):
        dst[i] = src[i]


@instr("{dst_data} = _mm512_maskz_loadu_ps(((1 << {N}) - 1), &{src_data});")
def mm512_maskz_loadu_ps(
    N: size,
    dst: [f32][16] @ AVX512,
    src: [f32][N] @ DRAM,
):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1
    assert N <= 16

    for i in seq(0, 16):
        if i < N:
            dst[i] = src[i]


@instr("_mm512_mask_storeu_ps(&{dst_data}, ((1 << {N}) - 1), {src_data});")
def mm512_mask_storeu_ps(N: size, dst: [f32][N] @ DRAM, src: [f32][16] @ AVX512):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1
    assert N <= 16

    for i in seq(0, 16):
        if i < N:
            dst[i] = src[i]


@instr("{C_data} = _mm512_fmadd_ps({A_data}, {B_data}, {C_data});")
def mm512_fmadd_ps(
    A: [f32][16] @ AVX512,
    B: [f32][16] @ AVX512,
    C: [f32][16] @ AVX512,
):
    assert stride(A, 0) == 1
    assert stride(B, 0) == 1
    assert stride(C, 0) == 1

    for i in seq(0, 16):
        C[i] += A[i] * B[i]


@instr(
    "{C_data} = _mm512_mask_fmadd_ps({A_data}, ((1 << {N}) - 1), {B_data}, {C_data});"
)
def mm512_mask_fmadd_ps(
    N: size,
    A: [f32][16] @ AVX512,
    B: [f32][16] @ AVX512,
    C: [f32][16] @ AVX512,
):
    assert N >= 1
    assert N < 16
    assert stride(A, 0) == 1
    assert stride(B, 0) == 1
    assert stride(C, 0) == 1

    for i in seq(0, 16):
        if i < N:
            C[i] += A[i] * B[i]


@instr("{dst_data} = _mm512_max_ps({src_data}, (__m512){{0}});")
def mm512_relu_ps(dst: [f32][16] @ AVX512, src: [f32][16] @ AVX512):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 16):
        dst[i] = relu(src[i])


# ---------------------------------------------------------------------------- #


# TODO: this "dumb" instruction exists as a hack around our current lack of
#  overcompute. We'll revisit the proper way of doing this post-deadline.


@instr("{dst_data} = _mm512_set1_ps({src_data});")
def mm512_mask_set1_ps(
    N: size,
    dst: [f32][16] @ AVX512,
    src: [f32][1],
):
    assert N >= 1
    assert N < 16
    assert stride(dst, 0) == 1

    for i in seq(0, 16):
        if i < N:
            dst[i] = src[0]


# ---------------------------------------------------------------------------- #


@instr("{dst_data} = _mm512_set1_ps({src_data});")
def mm512_set1_ps(
    dst: [f32][16] @ AVX512,
    src: [f32][1],
):
    assert stride(dst, 0) == 1

    for i in seq(0, 16):
        dst[i] = src[0]


# --------------------------------------------------------------------------- #
#   Complex AVX2 operations
# --------------------------------------------------------------------------- #


@instr("{out_data} = _mm256_xor_ps({out_data}, {out_data});")
def avx2_set0_ps(out: [f32][8] @ AVX2):
    assert stride(out, 0) == 1

    for i in seq(0, 8):
        out[i] = 0.0


@instr(
    """
{{
  __m256 ones = {{ 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f }};
  __m256 dst = _mm256_loadu_ps(&{dst_data});
  _mm256_storeu_ps(&{dst_data}, _mm256_fmadd_ps(ones, dst, {val_data}));
}}
"""
)
def avx2_fmadd_memu_ps(dst: [f32][8] @ DRAM, val: [f32][8] @ AVX2):
    assert stride(dst, 0) == 1
    assert stride(val, 0) == 1

    for i in seq(0, 8):
        dst[i] += val[i]


@instr(
    """
{out_data} = _mm256_blendv_ps ({z_data}, {y_data}, 
_mm256_cmp_ps ({x_data}, {v_data}, _CMP_LT_OQ));
"""
)
def avx2_select_ps(
    out: [f32][8] @ AVX2,
    x: [f32][8] @ AVX2,
    v: [f32][8] @ AVX2,
    y: [f32][8] @ AVX2,
    z: [f32][8] @ AVX2,
):
    # WARNING: This instruction above use a lower precision
    #    float32 (C float) than the implementation of
    #    the builtin which uses float64 (C double)
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(v, 0) == 1
    assert stride(y, 0) == 1
    assert stride(z, 0) == 1

    for i in seq(0, 8):
        out[i] = select(x[i], v[i], y[i], z[i])


@instr(
    """
{out_data} = _mm256_blendv_pd ({z_data}, {y_data},
_mm256_cmp_pd ({x_data}, {v_data}, _CMP_LT_OQ));
"""
)
def avx2_select_pd(
    out: [f64][4] @ AVX2,
    x: [f64][4] @ AVX2,
    v: [f64][4] @ AVX2,
    y: [f64][4] @ AVX2,
    z: [f64][4] @ AVX2,
):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(v, 0) == 1
    assert stride(y, 0) == 1
    assert stride(z, 0) == 1

    for i in seq(0, 4):
        out[i] = select(x[i], v[i], y[i], z[i])


@instr(
    """
    {{
        __m256 tmp = _mm256_hadd_ps({x_data}, {x_data});
        tmp = _mm256_hadd_ps(tmp, tmp);
        __m256 upper_bits = _mm256_castps128_ps256(_mm256_extractf128_ps(tmp, 1));
        tmp = _mm256_add_ps(tmp, upper_bits);
        *{result} += _mm256_cvtss_f32(tmp);
    }}
    """
)
def avx2_assoc_reduce_add_ps(x: [f32][8] @ AVX2, result: f32):
    # WARNING: This instruction assumes float addition associativity
    assert stride(x, 0) == 1
    for i in seq(0, 8):
        result += x[i]


@instr(
    """
    {{
        __m256d tmp = _mm256_hadd_pd({x_data}, {x_data});
        __m256d upper_bits = _mm256_castpd128_pd256(_mm256_extractf128_pd (tmp, 1));
        tmp = _mm256_add_pd(tmp, upper_bits);
        *{result} += _mm256_cvtsd_f64(tmp);
    }}
    """
)
def avx2_assoc_reduce_add_pd(x: [f64][4] @ AVX2, result: f64):
    # WARNING: This instruction assumes float addition associativity
    assert stride(x, 0) == 1
    for i in seq(0, 4):
        result += x[i]


@instr("{dst_data} = _mm256_mul_ps({src_data}, _mm256_set1_ps(-1.0f));")
def avx2_sign_ps(dst: [f32][8] @ AVX2, src: [f32][8] @ AVX2):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 8):
        dst[i] = -src[i]


@instr("{dst_data} = _mm256_mul_pd({src_data}, _mm256_set1_pd(-1.0f));")
def avx2_sign_pd(dst: [f64][4] @ AVX2, src: [f64][4] @ AVX2):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 4):
        dst[i] = -src[i]


@instr("{dst_data} = _mm256_add_ps({src_data}, {dst_data});")
def avx2_reduce_add_wide_ps(dst: [f32][8] @ AVX2, src: [f32][8] @ AVX2):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 8):
        dst[i] += src[i]


@instr("{dst_data} = _mm256_add_pd({src_data}, {dst_data});")
def avx2_reduce_add_wide_pd(dst: [f64][4] @ AVX2, src: [f64][4] @ AVX2):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 4):
        dst[i] += src[i]


# TODO: Hack for procedure aliasing issue, can be deleted once we have
#      better way of handling aliasing
@instr("{dst_data} = {src_data};")
def avx2_reg_copy_ps(dst: [f32][8] @ AVX2, src: [f32][8] @ AVX2):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 8):
        dst[i] = src[i]


@instr("{dst_data} = {src_data};")
def avx2_reg_copy_pd(dst: [f64][4] @ AVX2, src: [f64][4] @ AVX2):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 4):
        dst[i] = src[i]


@instr(
    "__m256i opaque = _mm256_set1_epi8((1<<{N}) - 1);\n"
    + "_mm256_maskstore_ps(&{dst_data}, opaque, {src_data});"
)
def avx2_mask_storeu_ps(N: size, dst: [f32][N] @ DRAM, src: [f32][8] @ AVX2):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1
    assert N <= 8

    for i in seq(0, 8):
        if i < N:
            dst[i] = src[i]


# TODO: This is hacked specifically for the 2D blur with kernel size 3. We should
# be able to support fast integer division instruction selection in a better way.
@instr(
    """
    {{
        {out_data} = _mm256_mulhi_epu16({x_data}, _mm256_set1_epi16(43691));
        {out_data} = _mm256_srli_epi16({out_data}, 1);
    }}
    """
)
def avx2_ui16_divide_by_3(out: [ui16][16] @ AVX2, x: [ui16][16] @ AVX2):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1

    for xii in seq(0, 16):
        out[xii] = x[xii] / 3.0


# --------------------------------------------------------------------------- #
#  Prefixed AVX2 Operations
# --------------------------------------------------------------------------- #


@instr(
    """
{{
__m256i indices = _mm256_set_epi32(7, 6, 5, 4, 3, 2, 1, 0);
__m256i prefix = _mm256_set1_epi32({bound});
__m256i cmp = _mm256_cmpgt_epi32(prefix, indices);
{dst_data} = _mm256_maskload_ps(&{src_data}, cmp);
}}
"""
)
def mm256_prefix_load_ps(dst: [f32][8] @ AVX2, src: [f32][8] @ DRAM, bound: size):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1
    assert bound <= 8
    for i in seq(0, 8):
        if i < bound:
            dst[i] = src[i]


@instr(
    """
{{
__m256i indices = _mm256_set_epi32(7, 6, 5, 4, 3, 2, 1, 0);
__m256i prefix = _mm256_set1_epi32({bound});
__m256i cmp = _mm256_cmpgt_epi32(prefix, indices);
_mm256_maskstore_ps(&{dst_data}, cmp, {src_data});
}}
"""
)
def mm256_prefix_store_ps(dst: [f32][8] @ DRAM, src: [f32][8] @ AVX2, bound: size):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1
    assert bound <= 8
    for i in seq(0, 8):
        if i < bound:
            dst[i] = src[i]


@instr(
    """
{{
__m256i indices = _mm256_set_epi32(7, 6, 5, 4, 3, 2, 1, 0);
__m256i prefix = _mm256_set1_epi32({bound});
__m256i cmp = _mm256_cmpgt_epi32(prefix, indices);
__m256 add = _mm256_add_ps({x_data}, {y_data});
{out_data} = _mm256_blendv_ps ({out_data}, add, _mm256_castsi256_ps(cmp));
}}
"""
)
def mm256_prefix_add_ps(
    out: [f32][8] @ AVX2, x: [f32][8] @ AVX2, y: [f32][8] @ AVX2, bound: size
):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(y, 0) == 1
    assert bound <= 8

    for i in seq(0, 8):
        if i < bound:
            out[i] = x[i] + y[i]


@instr(
    """
{{
__m256i indices = _mm256_set_epi32(7, 6, 5, 4, 3, 2, 1, 0);
__m256i prefix = _mm256_set1_epi32({bound});
__m256i cmp = _mm256_cmpgt_epi32(prefix, indices);
__m256 mul = _mm256_mul_ps({x_data}, {y_data});
{out_data} = _mm256_blendv_ps ({out_data}, mul, _mm256_castsi256_ps(cmp));
}}
"""
)
def mm256_prefix_mul_ps(
    out: [f32][8] @ AVX2, x: [f32][8] @ AVX2, y: [f32][8] @ AVX2, bound: size
):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(y, 0) == 1
    assert bound <= 8

    for i in seq(0, 8):
        if i < bound:
            out[i] = x[i] * y[i]


@instr(
    """
{{
__m256i indices = _mm256_set_epi32(7, 6, 5, 4, 3, 2, 1, 0);
__m256i prefix = _mm256_set1_epi32({bound});
__m256i cmp = _mm256_cmpgt_epi32(prefix, indices);
__m256 sub = _mm256_sub_ps({x_data}, {y_data});
{out_data} = _mm256_blendv_ps ({out_data}, sub, _mm256_castsi256_ps(cmp));
}}
"""
)
def mm256_prefix_sub_ps(
    out: [f32][8] @ AVX2, x: [f32][8] @ AVX2, y: [f32][8] @ AVX2, bound: size
):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(y, 0) == 1
    assert bound <= 8

    for i in seq(0, 8):
        if i < bound:
            out[i] = x[i] - y[i]


@instr(
    """
{{
__m256i indices = _mm256_set_epi32(7, 6, 5, 4, 3, 2, 1, 0);
__m256i prefix = _mm256_set1_epi32({bound});
__m256i cmp = _mm256_cmpgt_epi32(prefix, indices);
__m256 div = _mm256_div_ps({x_data}, {y_data});
{out_data} = _mm256_blendv_ps ({out_data}, div, _mm256_castsi256_ps(cmp));
}}
"""
)
def mm256_prefix_div_ps(
    out: [f32][8] @ AVX2, x: [f32][8] @ AVX2, y: [f32][8] @ AVX2, bound: size
):
    assert stride(out, 0) == 1
    assert stride(x, 0) == 1
    assert stride(y, 0) == 1
    assert bound <= 8

    for i in seq(0, 8):
        if i < bound:
            out[i] = x[i] / y[i]


@instr(
    """
{{
__m256i indices = _mm256_set_epi32(7, 6, 5, 4, 3, 2, 1, 0);
__m256i prefix = _mm256_set1_epi32({bound});
__m256i cmp = _mm256_cmpgt_epi32(prefix, indices);
__m256 bc = _mm256_broadcast_ss(&{val_data});
{out_data} = _mm256_blendv_ps ({out_data}, bc, _mm256_castsi256_ps(cmp));
}}
"""
)
def mm256_prefix_broadcast_ss(out: [f32][8] @ AVX2, val: [f32][1], bound: size):
    assert stride(out, 0) == 1
    assert bound <= 8

    for i in seq(0, 8):
        if i < bound:
            out[i] = val[0]


# --------------------------------------------------------------------------- #
#   f32 to f64 conversion
# --------------------------------------------------------------------------- #


@instr("{dst_data} = _mm256_cvtps_pd(_mm256_extractf128_ps({src_data}, 0));")
def avx2_convert_f32_lower_to_f64(dst: [f64][4] @ AVX2, src: [f32][8] @ AVX2):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 4):
        dst[i] = src[i]


@instr("{dst_data} = _mm256_cvtps_pd(_mm256_extractf128_ps({src_data}, 1));")
def avx2_convert_f32_upper_to_f64(dst: [f64][4] @ AVX2, src: [f32][8] @ AVX2):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 4):
        dst[i] = src[4 + i]



"""Query AST Classes

This module contains a reflection of Exo's internal AST structures.

They are organized into a class hierarchy of Python dataclasses as
follows.

QueryAST
  Proc  ( name : str, args : list[FnArg], assertions : list[Expr],
                      body : list[Stmt],  instruction : Optional[str] )
  FnArg ( name : str, type : Type, memory : Optional[Memory] )
  Stmt
    Assign    ( name : str,   lhs_type : Type,    idx : list[Expr],
                              rhs  : Expr )
    Reduce    ( name : str,   lhs_type : Type,    idx : list[Expr],
                              rhs  : Expr )
    WriteConfig ( config : Config, field : str,
                              rhs  : Expr )
    Pass      ()
    If        ( cond : Expr,  body : list[Stmt],  orelse : list[Stmt] )
    For       ( name : str,   lo   : Expr,        hi : Expr,
                              body : list[Stmt],  is_par : bool )
    Alloc     ( name : str,   type : Type,        memory : Optional[Memory] )
    Call      ( proc : str,   args : list[Expr] )
    WindowStmt( name : str,   rhs  : Expr )
  Expr
    Read    ( name : str,   idx  : list[Expr],    type : Type )
    Const   ( val  : Any,   type : Type )
    USub    ( arg  : Expr,  type : Type )
    BinOp   ( op   : str,   lhs  : Expr,
                            rhs  : Expr,          type : Type )
    BuiltIn ( func : str,   args : list[Expr],    type : Type )
    WindowExpr( name : str, idx : list[WAccess],  type : Type )
    StrideExpr( name : str, dim : int,            type : Type )
    ReadConfig( config : Config, field : str,     type : Type )
  WAccess
    Interval( lo : Expr, hi : Expr )
    Point( pt : Expr )
  Type
    R()
    f16()
    f32()
    f64()
    i8()
    i32()
    bool()
    int()
    index()
    size()
    stride()
    tensor( hi : list[Expr], is_window : bool, type : Type )

"""

from dataclasses import dataclass as _dataclass
from typing import Any as _Any
from typing import Optional as _Optional

from .configs import Config as _Config
from .memory import Memory as _Memory


# ----------------------------------------------------------------------
# -- base classes for all asts returned from the reflection interface --


class QueryAST:
    def __init__(self):
        raise Exception("Should never try to instantiate QueryAST")


class Type(QueryAST):
    def __init__(self):
        raise Exception("Should never try to instantiate Type")


class Expr(QueryAST):
    def __init__(self):
        raise Exception("Should never try to instantiate Expr")


class WAccess(QueryAST):
    def __init__(self):
        raise Exception("Should never try to instantiate WAccess")


class Stmt(QueryAST):
    def __init__(self):
        raise Exception("Should never try to instantiate Stmt")


# -----------------------------------
# -- QueryAST --> Type --> _______ --


@_dataclass
class R(Type):
    pass


@_dataclass
class f16(Type):
    pass


@_dataclass
class f32(Type):
    pass


@_dataclass
class f64(Type):
    pass


@_dataclass
class i8(Type):
    pass


@_dataclass
class i32(Type):
    pass


@_dataclass
class bool(Type):
    pass


@_dataclass
class int(Type):
    pass


@_dataclass
class index(Type):
    pass


@_dataclass
class size(Type):
    pass


@_dataclass
class stride(Type):
    pass


@_dataclass
class tensor(Type):
    hi: list[Expr]
    is_window: bool
    type: Type


# --------------------------------------
# -- QueryAST --> WAccess --> _______ --


@_dataclass
class Interval(WAccess):
    lo: Expr
    hi: Expr


@_dataclass
class Point(WAccess):
    pt: Expr


# -----------------------------------
# -- QueryAST --> Expr --> _______ --


@_dataclass
class Read(Expr):
    name: str
    idx: list[Expr]
    type: Type


@_dataclass
class Const(Expr):
    val: _Any
    type: Type


@_dataclass
class USub(Expr):
    arg: Expr
    type: Type


@_dataclass
class BinOp(Expr):
    op: str
    lhs: Expr
    rhs: Expr
    type: Type


@_dataclass
class BuiltIn(Expr):
    func: str
    args: list[Expr]
    type: Type


@_dataclass
class WindowExpr(Expr):
    name: str
    idx: list[WAccess]
    type: Type


@_dataclass
class StrideExpr(Expr):
    name: str
    dim: int
    type: Type


@_dataclass
class ReadConfig(Expr):
    config: _Config
    field: str
    type: Type


# -----------------------------------
# -- QueryAST --> Stmt --> _______ --


@_dataclass
class Assign(Stmt):
    name: str
    lhs_type: Type
    idx: list[Expr]
    rhs: Expr


@_dataclass
class Reduce(Stmt):
    name: str
    lhs_type: Type
    idx: list[Expr]
    rhs: Expr


@_dataclass
class WriteConfig(Stmt):
    config: _Config
    field: str
    rhs: Expr


@_dataclass
class Pass(Stmt):
    pass


@_dataclass
class If(Stmt):
    cond: Expr
    body: list[Expr]
    orelse: list[Expr]


@_dataclass
class For(Stmt):
    name: str
    lo: Expr
    hi: Expr
    body: list[Expr]
    is_par: bool


@_dataclass
class Alloc(Stmt):
    name: str
    type: Type
    memory: _Optional[_Memory]


@_dataclass
class Call(Stmt):
    proc: str
    args: list[Expr]


@_dataclass
class WindowStmt(Stmt):
    name: str
    rhs: Expr


# --------------------------
# -- QueryAST --> _______ --


@_dataclass
class FnArg(QueryAST):
    name: str
    type: Type
    memory: _Optional[_Memory]


@_dataclass
class Proc(QueryAST):
    name: str
    args: list[FnArg]
    assertions: list[Expr]
    body: list[Stmt]
    instruction: _Optional[str]



from __future__ import annotations

import pytest

from exo import proc, DRAM


# ------- Window related tests ---------


def test_input1():
    @proc
    def foo(dst1: f32[8] @ DRAM, src1: [f32][8] @ DRAM):
        assert stride(src1, 0) == 1
        assert stride(dst1, 0) == 1

        for i in seq(0, 8):
            dst1[i] = src1[i]


def test_input2():
    @proc
    def foo(dst2: [f32][8] @ DRAM, src2: f32[8] @ DRAM):
        assert stride(src2, 0) == 1
        assert stride(dst2, 0) == 1

        for i in seq(0, 8):
            dst2[i] = src2[i]


def test_input3():
    with pytest.raises(TypeError, match="Could not verify assertion"):

        @proc
        def foo(dst2: [f32][8] @ DRAM, src2: f32[8] @ DRAM):
            assert stride(src2, 0) == 1
            assert stride(dst2, 0) == 1

            for i in seq(0, 8):
                dst2[i] = src2[i]

        @proc
        def bar(x: [f32][8]):
            foo(x, x)


def test_input4():
    @proc
    def foo(dst2: [f32][8] @ DRAM, src2: f32[8] @ DRAM):
        assert stride(src2, 0) == 1
        assert stride(dst2, 0) == 1

        for i in seq(0, 8):
            dst2[i] = src2[i]

    @proc
    def bar(x: [f32][8], y: [f32][8]):
        assert stride(x, 0) == 1
        assert stride(y, 0) == 1
        foo(x, y)


def test_window(golden):
    @proc
    def window(
        n: size,
        m: size,
        src: [i8][n, m] @ DRAM,
        dst: [i8][n, 16] @ DRAM,
    ):
        assert n <= 16
        assert m <= 16

        for i in seq(0, n):
            for j in seq(0, m):
                dst[i, j] = src[i, j]

    assert window.c_code_str() == golden


def test_stride_assert(golden):
    @proc
    def stride_assert(
        n: size,
        m: size,
        src: [i8][n, m] @ DRAM,
        dst: [i8][n, 16] @ DRAM,
    ):
        assert n <= 16
        assert m <= 16
        assert stride(src, 1) == 1
        assert stride(dst, 0) == 16
        assert stride(dst, 1) == 1

        for i in seq(0, n):
            for j in seq(0, m):
                dst[i, j] = src[i, j]

    assert stride_assert.c_code_str() == golden


def test_window_stmt(golden):
    @proc
    def window_stmt(n: size, m: size, x: f32[n, m]):
        y = x[:, 0]
        z: f32[n]
        for i in seq(0, n):
            z[i] = y[i]

    assert window_stmt.c_code_str() == golden


def test_normalize(golden):
    @proc
    def dot(m: size, x: [f32][m], y: [f32][m], r: f32):
        r = 0.0
        for i in seq(0, m):
            r += x[i] * y[i]

    @proc
    def proj(n: size, m: size, x: f32[n, m], y: f32[m, n]):
        assert n > 4
        assert m > 4
        xy: f32
        y2: f32
        dot(m, x[1, :], y[:, 2], xy)
        dot(m, y[:, 3], x[2, :], y2)

    assert proj.c_code_str() == golden



from __future__ import annotations

import pytest

from exo import proc
from exo.stdlib.scheduling import *
from exo.platforms.x86 import *


def test_divide_loop():
    @proc
    def foo():
        for i in seq(0, 10):
            pass
            pass

    i_loop = foo.find_loop("i")
    body = i_loop.body()

    # Loop always forwards to the outermost loop of the main loop nest
    # Loop body forwards to the main loop nest's body
    foo_guard = divide_loop(foo, "i", 5, ["io", "ii"], tail="guard")
    assert foo_guard.forward(i_loop) == foo_guard.find_loop("io")
    assert foo_guard.forward(body).parent().parent() == foo_guard.find_loop("ii")

    foo_cut = divide_loop(foo, "i", 5, ["io", "ii"], tail="cut")
    assert foo_cut.forward(i_loop) == foo_cut.find_loop("io")
    assert foo_cut.forward(body).parent() == foo_cut.find_loop("ii")

    foo_cg = divide_loop(foo, "i", 5, ["io", "ii"], tail="cut_and_guard")
    assert foo_cg.forward(i_loop) == foo_cg.find_loop("io")
    assert foo_cg.forward(body).parent() == foo_cg.find_loop("ii")


def test_mult_loop():
    @proc
    def foo():
        for i in seq(0, 10):
            for j in seq(0, 30):
                pass
                pass

    i_loop = foo.find_loop("i")
    j_loop = foo.find_loop("j")
    body = j_loop.body()
    foo = mult_loops(foo, "i j", "ij")

    assert foo.forward(body) == foo.find_loop("ij").body()


def test_fuse_loops():
    @proc
    def foo(n: size, x: R[n]):
        assert n > 3
        y: R[n]
        for i in seq(3, n):
            y[i] = x[i]
        for j in seq(3, n):
            x[j] = y[j] + 1.0

    loop1 = foo.find_loop("i")
    body1 = loop1.body()
    loop2 = foo.find_loop("j")
    body2 = loop2.body()

    both_loops = loop1.expand(0, 1)

    foo = fuse(foo, loop1, loop2)

    # Loop bodies are consecutive
    assert foo.forward(body1)[-1].next() == foo.forward(body2)[0]

    # Block containing both loops forwards to block containing fused loop.
    assert foo.forward(both_loops) == foo.find_loop("i").as_block()
